"""Types used in pydantic fields."""
# ruff: noqa: F821
from __future__ import annotations

import functools
import inspect
from typing import Any, Dict, List, Literal, Protocol, Union, runtime_checkable

try:
    from pydantic.v1 import Field, StrictBool
    from pydantic.v1.fields import ModelField
    from pydantic.v1.schema import SkipField
except ImportError:  # pragma: no cov
    from pydantic import Field, StrictBool
    from pydantic.fields import ModelField
    from pydantic.schema import SkipField

from typing_extensions import Annotated, TypedDict

from vizro.charts._charts_utils import _DashboardReadyFigure


# Used to describe _DashboardReadyFigure so we can keep CapturedCallable generic rather than referring to
# _DashboardReadyFigure explicitly.
@runtime_checkable
class _SupportsCapturedCallable(Protocol):
    _captured_callable: CapturedCallable


class CapturedCallable:
    """Stores a captured function call to use in a dashboard.

    Users do not need to instantiate this class directly. Instances are instead generated automatically
    through the [`capture`][vizro.models.types.capture] decorator. Some of the functionality is similar to
    `functools.partial`.

    Ready-to-use `CapturedCallable` instances are provided by Vizro. In this case refer to the user guide on
    [Charts/Graph][graph], [Table][table] or [Actions][pre-defined-actions] to see available choices.

    (Advanced) In case you would like to create your own `CapturedCallable`, please refer to the user guide on
    [custom charts](../user_guides/custom_charts.md), [custom tables][custom-table] or
    [custom actions][custom-actions].
    """

    def __init__(self, function, /, *args, **kwargs):
        """Creates a new CapturedCallable object that will be able to re-run `function`.

        Partially binds *args and **kwargs to the function call.

        Raises:
            ValueError if `function` contains positional-only or variadic positional parameters (*args).
        """
        # It is difficult to get positional-only and variadic positional arguments working at the same time as
        # variadic keyword arguments. Ideally we would do the __call__ as
        # self.__function(*bound_arguments.args, **bound_arguments.kwargs) as in the
        # Python documentation. This would handle positional-only and variadic positional arguments better but makes
        # it more difficult to handle variadic keyword arguments due to https://bugs.python.org/issue41745.
        # Hence we abandon bound_arguments.args and bound_arguments.kwargs in favor of just using
        # self.__function(**bound_arguments.arguments).
        parameters = inspect.signature(function).parameters
        invalid_params = {
            param.name
            for param in parameters.values()
            if param.kind in [inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.VAR_POSITIONAL]
        }

        if invalid_params:
            raise ValueError(
                f"Invalid parameter {', '.join(invalid_params)}. CapturedCallable does not accept functions with "
                f"positional-only or variadic positional parameters (*args)."
            )

        self.__function = function
        self.__bound_arguments = inspect.signature(function).bind_partial(*args, **kwargs).arguments
        self.__unbound_arguments = [
            param for param in parameters.values() if param.name not in self.__bound_arguments
        ]  # Maintaining the same order here is important.

        # A function can only ever have one variadic keyword parameter. {""} is just here so that var_keyword_param
        # is always unpacking a one element set.
        (var_keyword_param,) = {
            param.name for param in parameters.values() if param.kind == inspect.Parameter.VAR_KEYWORD
        } or {""}

        # Since we do __call__ as self.__function(**bound_arguments.arguments), we need to restructure the arguments
        # a bit to put the kwargs in the right place.
        # For a function with parameter **kwargs this converts self.__bound_arguments = {"kwargs": {"a": 1}} into
        # self.__bound_arguments = {"a": 1}.
        if var_keyword_param in self.__bound_arguments:
            self.__bound_arguments.update(self.__bound_arguments[var_keyword_param])
            del self.__bound_arguments[var_keyword_param]

    def __call__(self, *args, **kwargs):
        """Run the `function` with the initially bound arguments overridden by `**kwargs`.

        *args are possible here, but cannot be used to override arguments bound in `__init__` - just to
        provide additional arguments. You can still override arguments that were originally given
        as positional using their argument name.
        """
        if args and kwargs:
            # In theory we could probably lift this restriction, but currently we don't need to and we'd need
            # to give careful thought on the right way to handle cases where there's ambiguity in the
            # self.__function call as the same argument is potentially being provided through both *args and **kwargs.
            raise ValueError("CapturedCallable does not support calling with both positional and keyword arguments.")

        # In order to avoid any ambiguity in the call to self.__function, we cannot provide use the *args directly.
        # Instead they must converted to keyword arguments and so we need to match them up with the right keywords.
        # Since positional-only or variadic positional parameters are not possible (they raise ValueError in __init__)
        # the only possible type of argument *args could be address is positional-or-keyword.
        if args:
            unbound_positional_arguments = [
                param.name
                for param in self.__unbound_arguments
                if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            ]
            if len(args) > len(unbound_positional_arguments):
                # TypeError to match the standard Python exception raised in this case.
                raise TypeError(
                    f"CapturedCallable takes {len(unbound_positional_arguments)} "
                    f"positional arguments but {len(args)} were given."
                )

            # No need to handle case that len(args) < len(unbound_positional_arguments),
            # since this will already raise error in the following function call.
            return self.__function(**dict(zip(unbound_positional_arguments, args)), **self.__bound_arguments)

        return self.__function(**{**self.__bound_arguments, **kwargs})

    def __getitem__(self, arg_name: str):
        """Gets the value of a bound argument."""
        return self.__bound_arguments[arg_name]

    def __delitem__(self, arg_name: str):
        """Deletes a bound argument."""
        del self.__bound_arguments[arg_name]

    @property
    def _arguments(self):
        # TODO: This is used twice: in _get_parametrized_config and in vm.Action and should be removed when those
        # references are removed.
        return self.__bound_arguments

    # TODO-actions: Find a way how to compare CapturedCallable and function
    @property
    def _function(self):
        return self.__function

    @classmethod
    def __get_validators__(cls):
        """Makes type compatible with pydantic model without needing arbitrary_types_allowed."""
        yield cls._parse_json

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any], field: ModelField):
        """Generates schema for field of this type."""
        raise SkipField(f"{cls.__name__} {field.name} is excluded from the schema.")

    @classmethod
    def _parse_json(
        cls,
        callable_config: Union[_SupportsCapturedCallable, CapturedCallable, Dict[str, Any]],
        field: ModelField,
    ) -> CapturedCallable:
        """Parses callable_config specification from JSON/YAML to a CapturedCallable.

        This uses the hydra syntax for _target_ but none of the other bits and we don't actually use hydra
        to implement it. In future, we might like to switch to using hydra's actual implementation
        which would allow nested functions (e.g. for transformers?) and to specify the path to a _target_ that lives
        outside of vizro.plotly_express. See https://hydra.cc/docs/advanced/instantiate_objects/overview/.
        """
        if isinstance(callable_config, CapturedCallable):
            # e.g. an action function that is already CapturedCallable
            return callable_config
        elif isinstance(callable_config, _SupportsCapturedCallable):
            # e.g. a _DashboardReadyFigure that has CapturedCallable in a property ._captured_callable
            return callable_config._captured_callable
        elif not isinstance(callable_config, dict):
            raise ValueError(
                "You must provide a valid CapturedCallable object. If you are using a plotly express figure, ensure "
                "that you are using `import vizro.plotly.express as px`. If you are using a table figure, make "
                "sure you are using `from vizro.tables import dash_data_table`. If you are using a custom figure or "
                "action, that your function uses the @capture decorator."
            )

        # Try to import function given in _target_ from the import_path property of the pydantic field.
        try:
            function_name = callable_config.pop("_target_")
        except KeyError as exc:
            raise ValueError(
                "CapturedCallable object must contain the key '_target_' that gives the target function."
            ) from exc

        import_path = field.field_info.extra["import_path"]
        try:
            function = getattr(import_path, function_name)
        except AttributeError as exc:
            raise ValueError(f"_target_={function_name} cannot be imported from {import_path.__name__}.") from exc

        # All the other items in figure are the keyword arguments to pass into function.
        function_kwargs = callable_config

        # It would seem natural to return cls(function, **function_kwargs) here, but the function is already decorated
        # with @capture, and so that would return a nested CapturedCallable.
        captured_callable = function(**function_kwargs)
        if isinstance(captured_callable, CapturedCallable):
            # e.g. an action function that is already CapturedCallable
            return captured_callable
        elif isinstance(captured_callable, _SupportsCapturedCallable):
            # e.g. a _DashboardReadyFigure that has CapturedCallable in a property ._captured_callable
            return captured_callable._captured_callable
        else:
            raise ValueError(f"_target_={function_name} must be wrapped in the @capture decorator.")


class capture:
    """Captures a function call to create a [`CapturedCallable`][vizro.models.types.CapturedCallable].

    This is used to add the functionality required to make graphs and actions work in a dashboard.
    Typically, it should be used as a function decorator. There are three possible modes: `"graph"`, `"table"` and
    `"action"`.

    Examples:
        >>> @capture("graph")
        >>> def graph_function():
        >>>     ...
        >>> @capture("table")
        >>> def table_function():
        >>>     ...
        >>> @capture("action")
        >>> def action_function():
        >>>     ...

    For further help on the use of `@capture("graph")`, you can refer to the guide on
    [custom graphs](../user_guides/custom_charts.md).
    For further help on the use of `@capture("table")`, you can refer to the guide on
    [custom tables](../user_guides/table#custom-table).
    For further help on the use of `@capture("action")`, you can refer to the guide on
    [custom actions](../user_guides/actions/#custom-actions).
    """

    def __init__(self, mode: Literal["graph", "action", "table"]):
        """Instantiates the decorator to capture a function call. Valid modes are "graph", "table" and "action"."""
        self._mode = mode

    def __call__(self, func, /):
        """Produces a CapturedCallable or _DashboardReadyFigure.

        mode="action" and mode="table" give a CapturedCallable, while mode="graph" gives a _DashboardReadyFigure that
        contains a CapturedCallable. In both cases, the CapturedCallable is based on func and the provided
        *args and **kwargs.
        """
        if self._mode == "graph":
            # The more difficult case, where we need to still have a valid plotly figure that renders in a notebook.
            # Hence we attach the CapturedCallable as a property instead of returning it directly.
            # TODO: move point of checking that data_frame argument exists earlier on.
            # TODO: also would be nice to raise errors in CapturedCallable.__init__ at point of function definition
            #  rather than point of calling if possible.
            @functools.wraps(func)
            def wrapped(*args, **kwargs) -> _DashboardReadyFigure:
                if "data_frame" not in inspect.signature(func).parameters:
                    raise ValueError(f"{func.__name__} must have data_frame argument to use capture('graph').")

                # We need to capture function upfront in order to find value of data_frame argument: since it could be
                # positional or keyword, this is much more robust than trying to get it out of arg or kwargs ourselves.
                captured_callable: CapturedCallable = CapturedCallable(func, *args, **kwargs)

                try:
                    captured_callable["data_frame"]
                except KeyError as exc:
                    raise ValueError(f"{func.__name__} must supply a value to data_frame argument.") from exc

                if isinstance(captured_callable["data_frame"], str):
                    # Enable running e.g. px.scatter("iris") from the Python API. Don't actually run the function
                    # because it won't get work as there's no data. It's vital we don't fetch data from the data manager
                    # yet either, because otherwise all lazy data will be loaded before the dashboard is started.
                    # This case is not relevant for the JSON/YAML API, which is handled separately through validation of
                    # CapturedCallable.
                    fig = _DashboardReadyFigure()
                else:
                    # Standard case for px.scatter(df: pd.DataFrame).
                    fig = func(*args, **kwargs)
                    fig.__class__ = _DashboardReadyFigure

                fig._captured_callable = captured_callable
                return fig

            return wrapped
        elif self._mode == "action":
            # The "normal" case where we just capture the function call.
            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                # Note this is basically the same as partial(func, *args, **kwargs)
                return CapturedCallable(func, *args, **kwargs)

            return wrapped
        elif self._mode == "table":

            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                if "data_frame" not in inspect.signature(func).parameters:
                    raise ValueError(f"{func.__name__} must have data_frame argument to use capture('table').")

                captured_callable: CapturedCallable = CapturedCallable(func, *args, **kwargs)

                try:
                    captured_callable["data_frame"]
                except KeyError as exc:
                    raise ValueError(f"{func.__name__} must supply a value to data_frame argument.") from exc
                return captured_callable

            return wrapped
        raise ValueError(
            "Valid modes of the capture decorator are @capture('graph'), @capture('action') or @capture('table')."
        )


# Types used for selector values and options. Note the docstrings here are rendered on the API reference.
SingleValueType = Union[StrictBool, float, str]
"""Permissible value types for single-value selectors. Values are displayed as default."""
MultiValueType = Union[List[StrictBool], List[float], List[str]]
"""Permissible value types for multi-value selectors. Values are displayed as default."""


class OptionsDictType(TypedDict):
    """Permissible sub-type for OptionsType. Needs to be in the format of {"label": XXX, "value": XXX}."""

    label: str
    value: SingleValueType


OptionsType = Union[List[StrictBool], List[float], List[str], List[OptionsDictType]]
"""Permissible options types for selectors. Options are available choices for user to select from."""

# All the below types rely on models and so must use ForwardRef (i.e. "Checklist" rather than actual Checklist class).
SelectorType = Annotated[
    Union["Checklist", "Dropdown", "RadioItems", "RangeSlider", "Slider"],
    Field(
        discriminator="type",
        description="Selectors to be used inside a control.",
    ),
]
"""Discriminated union. Type of selector to be used inside a control: [`Checklist`][vizro.models.Checklist],
[`Dropdown`][vizro.models.Dropdown], [`RadioItems`][vizro.models.RadioItems],
[`RangeSlider`][vizro.models.RangeSlider] or [`Slider`][vizro.models.Slider]."""

_FormComponentType = Annotated[
    Union[SelectorType, "Button", "UserInput"],
    Field(
        discriminator="type",
        description="Components that can be used to receive user input within a form.",
    ),
]

ControlType = Annotated[
    Union["Filter", "Parameter"],
    Field(
        discriminator="type",
        description="Control that affects components on the page.",
    ),
]
"""Discriminated union. Type of control that affects components on the page: [`Filter`][vizro.models.Filter] or
[`Parameter`][vizro.models.Parameter]."""

ComponentType = Annotated[
    Union["Button", "Card", "Graph", "Table"],
    Field(
        discriminator="type",
        description="Component that makes up part of the layout on the page.",
    ),
]
"""Discriminated union. Type of component that makes up part of the layout on the page:
[`Button`][vizro.models.Button], [`Card`][vizro.models.Card], [`Table`][vizro.models.Table] or
[`Graph`][vizro.models.Graph]."""

NavPagesType = Union[List[str], Dict[str, List[str]]]
"List of page IDs or a mapping from name of a group to a list of page IDs (for hierarchical sub-navigation)."

NavSelectorType = Annotated[
    Union["Accordion", "NavBar"],
    Field(discriminator="type", description="Component for rendering navigation."),
]
"""Discriminated union. Type of component for rendering navigation:
[`Accordion`][vizro.models.Accordion] or [`NavBar`][vizro.models.NavBar]."""



"""Unit tests for vizro.integrations.kedro."""
import types
from pathlib import Path

import pytest
import yaml

kedro = pytest.importorskip("kedro")

from kedro.io import DataCatalog  # noqa: E402

from vizro.integrations.kedro import datasets_from_catalog  # noqa: E402


@pytest.fixture
def catalog_path():
    return Path(__file__).parent / "fixtures/test_catalog.yaml"


def test_datasets_from_catalog(catalog_path):
    catalog = DataCatalog.from_config(yaml.safe_load(catalog_path.read_text(encoding="utf-8")))
    assert "companies" in datasets_from_catalog(catalog)
    assert isinstance(datasets_from_catalog(catalog), dict)
    assert isinstance(datasets_from_catalog(catalog)["companies"], types.MethodType)



import logging

import plotly.io as pio
from plotly import graph_objects as go

from vizro._themes import dark, light

pio.templates["vizro_dark"] = dark
pio.templates["vizro_light"] = light
pio.templates.default = "vizro_dark"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.info(
    "Overwriting plotly default template with vizro chart template. "
    "To revert back to the default template, run `import plotly.io as pio; pio.templates.default = 'plotly'`."
)


class _DashboardReadyFigure(go.Figure):
    # Just for IDE completion and to define new attribute
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        data = args[0] if args else None
        #  refers to the data argument of https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html
        if isinstance(data, dict):
            self._captured_callable = data.get("_captured_callable", None)

    # This fixes copy.deepcopy(_DashboardReadyFigure), but pickling fails in case of functions decorated with `capture`
    # https://stackoverflow.com/questions/52185507/pickle-and-decorated-classes-picklingerror-not-the-same-object
    # If wanting to save _DashboardReadyFigure objects, we currently recommend "dill" (see also example in link)
    # https://pypi.org/project/dill/
    def __reduce__(self):
        _, (props,) = super().__reduce__()
        props["_captured_callable"] = self._captured_callable
        return self.__class__, (props,)



"""Module containing the standard implementation of `dash_table.DataTable`."""
from collections import defaultdict
from typing import Any, Dict, Mapping

import pandas as pd
from dash import dash_table

from vizro.models.types import capture


def _set_defaults_nested(supplied: Mapping[str, Any], defaults: Mapping[str, Any]) -> Dict[str, Any]:
    supplied = defaultdict(dict, supplied)
    for default_key, default_value in defaults.items():
        if isinstance(default_value, Mapping):
            supplied[default_key] = _set_defaults_nested(supplied[default_key], default_value)
        else:
            supplied.setdefault(default_key, default_value)
    return dict(supplied)


@capture("table")
def dash_data_table(data_frame: pd.DataFrame, **kwargs):
    """Standard `dash_table.DataTable`."""
    defaults = {
        "columns": [{"name": i, "id": i} for i in data_frame.columns],
        "style_as_list_view": True,
        "style_data": {"border_bottom": "1px solid var(--border-subtle-alpha-01)", "height": "40px"},
        "style_header": {
            "border_bottom": "1px solid var(--state-overlays-selected-hover)",
            "border_top": "1px solid var(--main-container-bg-color)",
            "height": "32px",
        },
    }
    kwargs = _set_defaults_nested(kwargs, defaults)
    return dash_table.DataTable(data=data_frame.to_dict("records"), **kwargs)



"""Pre-defined action function "_parameter" to be reused in `action` parameter of VizroBaseModels."""

from typing import Any, Dict, List

from dash import ctx

from vizro.actions._actions_utils import (
    _get_modified_page_figures,
)
from vizro.managers._model_manager import ModelID
from vizro.models.types import capture


@capture("action")
def _parameter(targets: List[str], **inputs: Dict[str, Any]) -> Dict[str, Any]:
    """Modifies parameters of targeted charts/components on page.

    Args:
        targets: List of target component ids to change parameters of.
        inputs: Dict mapping action function names with their inputs e.g.
            inputs = {'filters': [], 'parameters': ['gdpPercap'], 'filter_interaction': [], 'theme_selector': True}

    Returns:
        Dict mapping target component ids to modified charts/components e.g. {'my_scatter': Figure({})}
    """
    target_ids: List[ModelID] = [target.split(".")[0] for target in targets]  # type: ignore[misc]

    return _get_modified_page_figures(
        targets=target_ids,
        ctds_filter=ctx.args_grouping["external"]["filters"],
        ctds_filter_interaction=ctx.args_grouping["external"]["filter_interaction"],
        ctds_parameters=ctx.args_grouping["external"]["parameters"],
    )



import logging
from functools import wraps
from typing import List

import numpy as np

from vizro._constants import EMPTY_SPACE_CONST

logger = logging.getLogger(__name__)


def _log_call(method):
    @wraps(method)
    def _wrapper(self, *args, **kwargs):
        # We need to run method before logging so that @_log_call works for __init__.
        return_value = method(self, *args, **kwargs)
        logger.debug("Running %s.%s for model with id %s", self.__class__.__name__, method.__name__, self.id)
        return return_value

    return _wrapper


def get_unique_grid_component_ids(grid: List[List[int]]):
    unique_grid_idx = np.unique(grid)
    unique_grid_comp_idx = unique_grid_idx[unique_grid_idx != EMPTY_SPACE_CONST]
    return unique_grid_comp_idx



from __future__ import annotations

import itertools
from typing import TypedDict

from dash import html

from vizro.managers import model_manager


def _validate_pages(pages):
    """Reusable validator to check if provided Page IDs exist as registered pages."""
    from vizro.models import Page

    pages_as_list = list(itertools.chain(*pages.values())) if isinstance(pages, dict) else pages
    # Ideally we would use dash.page_registry or maybe dashboard.pages here, but we only register pages in
    # dashboard.pre_build and model manager cannot find a Dashboard at validation time.
    # page[0] gives the page model ID.
    registered_pages = [page[0] for page in model_manager._items_with_type(Page)]

    if not pages_as_list:
        raise ValueError("Ensure this value has at least 1 item.")

    if unknown_pages := [page for page in pages_as_list if page not in registered_pages]:
        raise ValueError(f"Unknown page ID {unknown_pages} provided to argument 'pages'.")
    return pages


# This is just used for type checking. Ideally it would inherit from some dash.development.base_component.Component
# (e.g. html.Div) as well as TypedDict, but that's not possible, and Dash does not have typing support anyway. When
# this type is used, the object is actually still a dash.development.base_component.Component, but this makes it easier
# to see what contract the component fulfills by making the expected keys explicit.
_NavBuildType = TypedDict("_NavBuildType", {"nav-bar": html.Div, "nav-panel": html.Div})



import logging
from typing import List, Literal

from dash import dash_table, dcc, html
from pandas import DataFrame

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator

import vizro.tables as vt
from vizro.managers import data_manager
from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components._components_utils import _process_callable_data_frame
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


class Table(VizroBaseModel):
    """Wrapper for table components to visualize in dashboard.

    Args:
        type (Literal["table"]): Defaults to `"table"`.
        figure (CapturedCallable): Table like object to be displayed. Current choices include:
            [`dash_table.DataTable`](https://dash.plotly.com/datatable).
        title (str): Title of the table. Defaults to `""`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["table"] = "table"
    figure: CapturedCallable = Field(..., import_path=vt, description="Table to be visualized on dashboard")
    title: str = Field("", description="Title of the table")
    actions: List[Action] = []

    _callable_object_id: str = PrivateAttr()

    # Component properties for actions and interactions
    _output_property: str = PrivateAttr("children")

    # validator
    set_actions = _action_validator_factory("active_cell")
    _validate_callable = validator("figure", allow_reuse=True, always=True)(_process_callable_data_frame)

    # Convenience wrapper/syntactic sugar.
    def __call__(self, **kwargs):
        kwargs.setdefault("data_frame", data_manager._get_component_data(self.id))
        return self.figure(**kwargs)

    # Convenience wrapper/syntactic sugar.
    def __getitem__(self, arg_name: str):
        # pydantic discriminated union validation seems to try Table["type"], which throws an error unless we
        # explicitly redirect it to the correct attribute.
        if arg_name == "type":
            return self.type
        return self.figure[arg_name]

    @_log_call
    def pre_build(self):
        if self.actions:
            kwargs = self.figure._arguments.copy()

            # This workaround is needed because the underlying table object requires a data_frame
            kwargs["data_frame"] = DataFrame()

            # The underlying table object is pre-built, so we can fetch its ID.
            underlying_table_object = self.figure._function(**kwargs)

            if not hasattr(underlying_table_object, "id"):
                raise ValueError(
                    "Underlying `Table` callable has no attribute 'id'. To enable actions triggered by the `Table`"
                    " a valid 'id' has to be provided to the `Table` callable."
                )

            self._callable_object_id = underlying_table_object.id

    def build(self):
        return dcc.Loading(
            html.Div(
                [
                    html.H3(self.title, className="table-title") if self.title else None,
                    html.Div(
                        dash_table.DataTable(**({"id": self._callable_object_id} if self.actions else {})), id=self.id
                    ),
                ],
                className="table-container",
                id=f"{self.id}_outer",
            ),
            color="grey",
            parent_className="loading-container",
        )



"""Contains utilities to create the action_callback_mapping."""

from itertools import chain
from typing import Any, Callable, Dict, List, NamedTuple

from dash import Output, State, dcc

from vizro.actions import _on_page_load, _parameter, export_data, filter_interaction
from vizro.managers import data_manager, model_manager
from vizro.managers._model_manager import ModelID
from vizro.models import Action, Page, Table, VizroBaseModel
from vizro.models._action._actions_chain import ActionsChain
from vizro.models._controls import Filter, Parameter
from vizro.models.types import ControlType


class ModelActionsChains(NamedTuple):
    model: VizroBaseModel
    actions_chains: List[ActionsChain]


def _get_actions(model) -> List[ActionsChain]:
    """Gets the list of trigger action chains in the `action` parameter for any model."""
    if hasattr(model, "selector"):
        return model.selector.actions
    elif hasattr(model, "actions"):
        return model.actions
    return []


def _get_all_actions_chains_on_page(page: Page) -> chain:  # type: ignore[type-arg]
    """Creates an itertools chain of all ActionsChains present on selected Page."""
    return chain(*(_get_actions(page_item) for page_item in chain([page], page.components, page.controls)))


def _get_model_actions_chains_mapping(page: Page) -> Dict[str, ModelActionsChains]:
    """Creates a mapping of model ids and ModelActionsChains for selected Page."""
    model_actions_chains_mapping = {}
    for page_item in chain([page], page.components, page.controls):
        model_actions_chains_mapping[page_item.id] = ModelActionsChains(
            model=page_item, actions_chains=_get_actions(page_item)
        )
    return model_actions_chains_mapping


def _get_triggered_page(action_id: ModelID) -> Page:  # type: ignore[return]
    """Gets the page where the provided `action_id` has been triggered."""
    for _, page in model_manager._items_with_type(Page):
        if any(
            action.id == action_id
            for actions_chain in _get_all_actions_chains_on_page(page)
            for action in actions_chain.actions
        ):
            return page


def _get_triggered_model(action_id: ModelID) -> VizroBaseModel:  # type: ignore[return]
    """Gets the model where the provided `action_id` has been triggered."""
    for _, page in model_manager._items_with_type(Page):
        for model_id, model_actions_chains in _get_model_actions_chains_mapping(page).items():
            if any(
                action.id == action_id
                for actions_chain in model_actions_chains.actions_chains
                for action in actions_chain.actions
            ):
                return model_actions_chains.model


def _get_components_with_data(action_id: ModelID) -> List[str]:
    """Gets all components that have a registered dataframe on the page where `action_id` was triggered."""
    page = _get_triggered_page(action_id=action_id)
    return [component.id for component in page.components if data_manager._has_registered_data(component.id)]


def _get_matching_actions_by_function(page: Page, action_function: Callable[[Any], Dict[str, Any]]) -> List[Action]:
    """Gets list of Actions on triggered page that match the provided action function."""
    return [
        action
        for actions_chain in _get_all_actions_chains_on_page(page)
        for action in actions_chain.actions
        if action.function._function == action_function
    ]


# CALLBACK STATES --------------
def _get_inputs_of_controls(action_id: ModelID, control_type: ControlType) -> List[State]:
    """Gets list of States for selected control_type of triggered page."""
    page = _get_triggered_page(action_id=action_id)
    return [
        State(
            component_id=control.selector.id,
            component_property=control.selector._input_property,
        )
        for control in page.controls
        if isinstance(control, control_type)
    ]


def _get_inputs_of_figure_interactions(
    action_id: ModelID, action_function: Callable[[Any], Dict[str, Any]]
) -> List[Dict[str, State]]:
    """Gets list of States for selected chart interaction `action_name` of triggered page."""
    figure_interactions_on_page = _get_matching_actions_by_function(
        page=_get_triggered_page(action_id=action_id),
        action_function=action_function,
    )
    inputs = []
    for action in figure_interactions_on_page:
        # TODO: Consider do we want to move the following logic into Model implementation
        triggered_model = _get_triggered_model(action_id=ModelID(str(action.id)))
        if isinstance(triggered_model, Table):
            inputs.append(
                {
                    "active_cell": State(
                        component_id=triggered_model._callable_object_id, component_property="active_cell"
                    ),
                    "derived_viewport_data": State(
                        component_id=triggered_model._callable_object_id,
                        component_property="derived_viewport_data",
                    ),
                }
            )
        else:
            inputs.append(
                {
                    "clickData": State(component_id=triggered_model.id, component_property="clickData"),
                }
            )

    return inputs


def _get_action_callback_inputs(action_id: ModelID) -> Dict[str, Any]:
    """Creates mapping of pre-defined action names and a list of States."""
    action_function = model_manager[action_id].function._function

    if action_function == export_data.__wrapped__:
        include_inputs = ["filters", "filter_interaction"]
    else:
        include_inputs = ["filters", "parameters", "filter_interaction", "theme_selector"]

    action_input_mapping = {
        "filters": (
            _get_inputs_of_controls(action_id=action_id, control_type=Filter) if "filters" in include_inputs else []
        ),
        "parameters": (
            _get_inputs_of_controls(action_id=action_id, control_type=Parameter)
            if "parameters" in include_inputs
            else []
        ),
        # TODO: Probably need to adjust other inputs to follow the same structure List[Dict[str, State]]
        "filter_interaction": (
            _get_inputs_of_figure_interactions(action_id=action_id, action_function=filter_interaction.__wrapped__)
            if "filter_interaction" in include_inputs
            else []
        ),
        "theme_selector": State("theme_selector", "on") if "theme_selector" in include_inputs else [],
    }
    return action_input_mapping


# CALLBACK OUTPUTS --------------
def _get_action_callback_outputs(action_id: ModelID) -> Dict[str, Output]:
    """Creates mapping of target names and their Output."""
    action_function = model_manager[action_id].function._function

    # The right solution for mypy here is to not e.g. define new attributes on the base but instead to get mypy to
    # recognize that model_manager[action_id] is of type Action and hence has the function attribute.
    # Ideally model_manager.__getitem__ would handle this itself, possibly with suitable use of a cast.
    # If not then we can do the cast to Action at the point of consumption here to avoid needing mypy ignores.

    try:
        targets = model_manager[action_id].function["targets"]
    except KeyError:
        targets = []

    if action_function == _parameter.__wrapped__:
        targets = [target.split(".")[0] for target in targets]

    if action_function == _on_page_load.__wrapped__:
        targets = _get_components_with_data(action_id=action_id)

    return {
        target: Output(
            component_id=target,
            component_property=model_manager[target]._output_property,
            allow_duplicate=True,
        )
        for target in targets
    }


def _get_export_data_callback_outputs(action_id: ModelID) -> Dict[str, List[State]]:
    """Gets mapping of relevant output target name and Outputs for `export_data` action."""
    action = model_manager[action_id]

    try:
        targets = action.function["targets"]
    except KeyError:
        targets = None

    if not targets:
        targets = _get_components_with_data(action_id=action_id)

    return {
        f"download_dataframe_{target}": Output(
            component_id={
                "type": "download_dataframe",
                "action_id": action_id,
                "target_id": target,
            },
            component_property="data",
        )
        for target in targets
    }


# CALLBACK COMPONENTS --------------
def _get_export_data_callback_components(action_id: ModelID) -> List[dcc.Download]:
    """Creates dcc.Downloads for target components of the `export_data` action."""
    action = model_manager[action_id]

    try:
        targets = action.function["targets"]
    except KeyError:
        targets = None

    if not targets:
        targets = _get_components_with_data(action_id=action_id)

    return [
        dcc.Download(
            id={
                "type": "download_dataframe",
                "action_id": action_id,
                "target_id": target,
            },
        )
        for target in targets
    ]



from typing import List, Literal

import dash_bootstrap_components as dbc
from dash import html

try:
    from pydantic.v1 import Field
except ImportError:  # pragma: no cov
    from pydantic import Field

from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._models_utils import _log_call


class UserInput(VizroBaseModel):
    """Component provided to `Form` to allow user text input.

    Args:
        type (Literal["user_input"]): Defaults to `"user_input"`.
        placeholder (str): Default text to display in input field. Defaults to `""`.
        actions (Optional[List[Action]]): Defaults to `[]`.
        title (str): Title to be displayed. Defaults to `""`.
        input_type (Literal["text", "number", "password", "email", "search", "tel", "url", "range", "hidden"]):
            Type of value to validate user input against. Defaults to `"text"`.
    """

    type: Literal["user_input"] = "user_input"
    title: str = Field("", description="Title to be displayed")
    placeholder: str = Field("", description="Default text to display in input field")
    input_type: Literal["text", "number", "password", "email", "search", "tel", "url", "range", "hidden"] = Field(
        "text", description="Type of value to validate user input against."
    )
    actions: List[Action] = []

    # Re-used validators
    _set_actions = _action_validator_factory("value")

    @_log_call
    def build(self):
        return html.Div(
            [
                html.P(self.title) if self.title else None,
                dbc.Input(
                    id=self.id,
                    placeholder=self.placeholder,
                    type=self.input_type,
                    persistence=True,
                    persistence_type="session",
                    debounce=True,
                    className="user_input",
                ),
            ],
            className="selector_container",
            id=f"{self.id}_outer",
        )



import pytest

from vizro.models.types import capture


@pytest.fixture
def identity_action_function():
    @capture("action")
    def _identity_action_function(arg=None):
        return arg

    return _identity_action_function



"""Unit tests for vizro.models.Table."""
import json

import plotly
import pytest
from dash import dash_table, dcc, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
import vizro.plotly.express as px
from vizro.actions import filter_interaction
from vizro.managers import data_manager
from vizro.models._action._action import Action
from vizro.tables import dash_data_table


@pytest.fixture
def dash_table_with_arguments():
    return dash_data_table(data_frame=px.data.gapminder(), style_header={"border": "1px solid green"})


@pytest.fixture
def dash_table_with_str_dataframe():
    return dash_data_table(data_frame="gapminder")


@pytest.fixture
def expected_table():
    return dcc.Loading(
        html.Div(
            [
                None,
                html.Div(dash_table.DataTable(), id="text_table"),
            ],
            className="table-container",
            id="text_table_outer",
        ),
        color="grey",
        parent_className="loading-container",
    )


@pytest.fixture
def expected_table_with_id():
    return dcc.Loading(
        html.Div(
            [
                None,
                html.Div(dash_table.DataTable(id="underlying_table_id"), id="text_table"),
            ],
            className="table-container",
            id="text_table_outer",
        ),
        color="grey",
        parent_className="loading-container",
    )


@pytest.fixture
def filter_interaction_action():
    return vm.Action(function=filter_interaction())


class TestDunderMethodsTable:
    def test_create_graph_mandatory_only(self, standard_dash_table):
        table = vm.Table(figure=standard_dash_table)

        assert hasattr(table, "id")
        assert table.type == "table"
        assert table.figure == standard_dash_table
        assert table.actions == []

    @pytest.mark.parametrize("id", ["id_1", "id_2"])
    def test_create_table_mandatory_and_optional(self, standard_dash_table, id):
        table = vm.Table(
            figure=standard_dash_table,
            id=id,
            actions=[],
        )

        assert table.id == id
        assert table.type == "table"
        assert table.figure == standard_dash_table

    def test_mandatory_figure_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Table()

    def test_failed_table_with_no_captured_callable(self, standard_go_chart):
        with pytest.raises(ValidationError, match="must provide a valid CapturedCallable object"):
            vm.Table(
                figure=standard_go_chart,
            )

    @pytest.mark.xfail(reason="This test is failing as we are not yet detecting different types of captured callables")
    def test_failed_table_with_wrong_captured_callable(self, standard_px_chart):
        with pytest.raises(ValidationError, match="must provide a valid table function vm.Table"):
            vm.Table(
                figure=standard_px_chart,
            )

    def test_getitem_known_args(self, dash_table_with_arguments):
        table = vm.Table(figure=dash_table_with_arguments)
        assert table["style_header"] == {"border": "1px solid green"}
        assert table["type"] == "table"

    def test_getitem_unknown_args(self, standard_dash_table):
        table = vm.Table(figure=standard_dash_table)
        with pytest.raises(KeyError):
            table["unknown_args"]

    def test_set_action_via_validator(self, standard_dash_table, identity_action_function):
        table = vm.Table(figure=standard_dash_table, actions=[Action(function=identity_action_function())])
        actions_chain = table.actions[0]
        assert actions_chain.trigger.component_property == "active_cell"


class TestProcessTableDataFrame:
    def test_process_figure_data_frame_str_df(self, dash_table_with_str_dataframe, gapminder):
        data_manager["gapminder"] = gapminder
        table_with_str_df = vm.Table(
            id="table",
            figure=dash_table_with_str_dataframe,
        )
        assert data_manager._get_component_data("table").equals(gapminder)
        assert table_with_str_df["data_frame"] == "gapminder"

    def test_process_figure_data_frame_df(self, standard_dash_table, gapminder):
        table_with_str_df = vm.Table(
            id="table",
            figure=standard_dash_table,
        )
        assert data_manager._get_component_data("table").equals(gapminder)
        with pytest.raises(KeyError, match="'data_frame'"):
            table_with_str_df.figure["data_frame"]


class TestPreBuildTable:
    def test_pre_build_no_actions_no_underlying_table_id(self, standard_dash_table):
        table = vm.Table(
            id="text_table",
            figure=standard_dash_table,
        )
        table.pre_build()

        assert hasattr(table, "_callable_object_id") is False

    def test_pre_build_actions_no_underlying_table_id_exception(self, standard_dash_table, filter_interaction_action):
        table = vm.Table(
            id="text_table",
            figure=standard_dash_table,
            actions=[filter_interaction_action],
        )
        with pytest.raises(ValueError, match="Underlying `Table` callable has no attribute 'id'"):
            table.pre_build()

    def test_pre_build_actions_underlying_table_id(self, dash_data_table_with_id, filter_interaction_action):
        table = vm.Table(
            id="text_table",
            figure=dash_data_table_with_id,
            actions=[filter_interaction_action],
        )
        table.pre_build()

        assert table._callable_object_id == "underlying_table_id"


class TestBuildTable:
    def test_table_build_mandatory_only(self, standard_dash_table, expected_table):
        table = vm.Table(
            id="text_table",
            figure=standard_dash_table,
        )

        table.pre_build()

        result = json.loads(json.dumps(table.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_table, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    def test_table_build_with_id(self, dash_data_table_with_id, filter_interaction_action, expected_table_with_id):
        table = vm.Table(
            id="text_table",
            figure=dash_data_table_with_id,
            actions=[filter_interaction_action],
        )

        table.pre_build()

        result = json.loads(json.dumps(table.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_table_with_id, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected



import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
import vizro.plotly.express as px
from vizro._constants import PARAMETER_ACTION_PREFIX
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def target_scatter_parameter_y(request, gapminder_2007, scatter_params):
    y = request.param
    scatter_params["y"] = y
    return px.scatter(gapminder_2007, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_scatter_parameter_hover_data(request, gapminder_2007, scatter_params):
    hover_data = request.param
    scatter_params["hover_data"] = hover_data
    return px.scatter(gapminder_2007, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_box_parameter_y(request, gapminder_2007, box_params):
    y = request.param
    box_params["y"] = y
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24)


@pytest.fixture
def target_scatter_parameter_y_and_x(request, gapminder_2007, scatter_params):
    y, x = request.param
    scatter_params["y"] = y
    scatter_params["x"] = x
    return px.scatter(gapminder_2007, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_box_parameter_y_and_x(request, gapminder_2007, box_params):
    y, x = request.param
    box_params["y"] = y
    box_params["x"] = x
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24)


@pytest.fixture
def callback_context_parameter_y(request):
    """Mock dash.callback_context that represents y-axis Parameter value selection."""
    y = request.param
    mock_callback_context = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [],
                "parameters": [
                    CallbackTriggerDict(
                        id="y_parameter",
                        property="value",
                        value=y,
                        str_id="y_parameter",
                        triggered=False,
                    )
                ],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="on",
                    value=True,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.fixture
def callback_context_parameter_hover_data(request):
    """Mock dash.callback_context that represents hover_data Parameter value selection."""
    hover_data = request.param
    mock_callback_context = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [],
                "parameters": [
                    CallbackTriggerDict(
                        id="hover_data_parameter",
                        property="value",
                        value=hover_data,
                        str_id="hover_data_parameter",
                        triggered=False,
                    )
                ],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="on",
                    value=True,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.fixture
def callback_context_parameter_y_and_x(request):
    """Mock dash.callback_context that represents y-axis Parameter value selection."""
    y, x = request.param
    mock_callback_context = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [],
                "parameters": [
                    CallbackTriggerDict(
                        id="y_parameter",
                        property="value",
                        value=y,
                        str_id="y_parameter",
                        triggered=False,
                    ),
                    CallbackTriggerDict(
                        id="x_parameter",
                        property="value",
                        value=x,
                        str_id="x_parameter",
                        triggered=False,
                    ),
                ],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="on",
                    value=True,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
class TestParameter:
    @pytest.mark.parametrize(
        "callback_context_parameter_y, target_scatter_parameter_y",
        [("pop", "pop"), ("gdpPercap", "gdpPercap"), ("NONE", None)],
        indirect=True,
    )
    def test_one_parameter_one_target(
        self,
        callback_context_parameter_y,
        target_scatter_parameter_y,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()
        expected = {
            "scatter_chart": target_scatter_parameter_y,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_parameter_hover_data, target_scatter_parameter_hover_data",
        [
            (["NONE"], [None]),
            (["NONE", "pop"], ["pop"]),
            (["NONE", "NONE"], [None]),
            (["ALL"], ["lifeExp", "pop", "gdpPercap"]),
            (["ALL", "lifeExp"], ["lifeExp", "pop", "gdpPercap"]),
            (["ALL", "NONE"], ["lifeExp", "pop", "gdpPercap"]),
        ],
        indirect=True,
    )
    def test_one_parameter_one_target_NONE_list(
        self,
        callback_context_parameter_hover_data,
        target_scatter_parameter_hover_data,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.hover_data"],
            selector=vm.Dropdown(
                id="hover_data_parameter", multi=True, options=["NONE", "lifeExp", "pop", "gdpPercap"], value=["NONE"]
            ),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()
        expected = {
            "scatter_chart": target_scatter_parameter_hover_data,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_parameter_y, target_scatter_parameter_y, target_box_parameter_y",
        [("pop", "pop", "pop"), ("gdpPercap", "gdpPercap", "gdpPercap")],
        indirect=True,
    )
    def test_one_parameter_multiple_targets(
        self,
        callback_context_parameter_y,
        target_scatter_parameter_y,
        target_box_parameter_y,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.y", "box_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()
        expected = {
            "scatter_chart": target_scatter_parameter_y,
            "box_chart": target_box_parameter_y,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_parameter_y_and_x, target_scatter_parameter_y_and_x",
        [(["pop", "continent"], ["pop", "continent"]), (["gdpPercap", "country"], ["gdpPercap", "country"])],
        indirect=True,
    )
    def test_multiple_parameters_one_target(
        self,
        callback_context_parameter_y_and_x,
        target_scatter_parameter_y_and_x,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter_x",
            targets=["scatter_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            id="test_parameter_y",
            targets=["scatter_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_x"].function()
        expected = {
            "scatter_chart": target_scatter_parameter_y_and_x,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x",
        [
            (["pop", "continent"], ["pop", "continent"], ["pop", "continent"]),
            (["gdpPercap", "country"], ["gdpPercap", "country"], ["gdpPercap", "country"]),
        ],
        indirect=True,
    )
    def test_multiple_parameters_multiple_targets(
        self,
        callback_context_parameter_y_and_x,
        target_scatter_parameter_y_and_x,
        target_box_parameter_y_and_x,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter_x",
            targets=["scatter_chart.y", "box_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            id="test_parameter_y",
            targets=["scatter_chart.x", "box_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_x"].function()
        expected = {
            "scatter_chart": target_scatter_parameter_y_and_x,
            "box_chart": target_box_parameter_y_and_x,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x",
        [
            (["pop", "continent"], ["pop", "pop"], ["continent", "continent"]),
            (["gdpPercap", "country"], ["gdpPercap", "gdpPercap"], ["country", "country"]),
        ],
        indirect=True,
    )
    def test_one_parameter_per_target_multiple_attributes(
        self,
        callback_context_parameter_y_and_x,
        target_scatter_parameter_y_and_x,
        target_box_parameter_y_and_x,
    ):
        # Creating and adding a Parameter object to the existing Page
        scatter_parameter = vm.Parameter(
            id="test_parameter_scatter",
            targets=["scatter_chart.y", "scatter_chart.x"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        box_parameter = vm.Parameter(
            id="test_parameter_box",
            targets=["box_chart.y", "box_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [scatter_parameter, box_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        scatter_parameter.pre_build()
        box_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_scatter"].function()
        expected = {
            "scatter_chart": target_scatter_parameter_y_and_x,
        }

        assert result == expected

        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_box"].function()
        expected = {
            "box_chart": target_box_parameter_y_and_x,
        }

        assert result == expected



"""Pre-defined action function "filter_interaction" to be reused in `action` parameter of VizroBaseModels."""

from typing import Any, Dict, List, Optional

from dash import ctx

from vizro.actions._actions_utils import (
    _get_modified_page_figures,
)
from vizro.managers._model_manager import ModelID
from vizro.models.types import capture


@capture("action")
def filter_interaction(
    targets: Optional[List[ModelID]] = None,
    **inputs: Dict[str, Any],
) -> Dict[str, Any]:
    """Filters targeted charts/components on page by clicking on data points or table cells of the source chart.

    To set up filtering on specific columns of the target graph(s), include these columns in the 'custom_data'
    parameter of the source graph e.g. `px.bar(..., custom_data=["species", "sepal_length"])`.
    If the filter interaction source is a table e.g. `vm.Table(..., actions=[filter_interaction])`,
    then the table doesn't need to have a 'custom_data' parameter set up.

    Args:
        targets: List of target component ids to filter by chart interaction. If missing, will target all valid
            components on page. Defaults to `None`.
        inputs: Dict mapping action function names with their inputs e.g.
            inputs = {'filters': [], 'parameters': ['gdpPercap'], 'filter_interaction': [], 'theme_selector': True}

    Returns:
        Dict mapping target component ids to modified charts/components e.g. {'my_scatter': Figure({})}
    """
    return _get_modified_page_figures(
        targets=targets,
        ctds_filter=ctx.args_grouping["external"]["filters"],
        ctds_filter_interaction=ctx.args_grouping["external"]["filter_interaction"],
        ctds_parameters=ctx.args_grouping["external"]["parameters"],
    )



"""Unit tests for vizro.models.Accordion."""
import re

import dash_bootstrap_components as dbc
import pytest
from asserts import assert_component_equal
from dash import html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
from vizro._constants import ACCORDION_DEFAULT_TITLE

pytestmark = pytest.mark.usefixtures("prebuilt_two_page_dashboard")


class TestAccordionInstantiation:
    """Tests accordion model instantiation."""

    def test_mandatory_only(self):
        accordion = vm.Accordion()

        assert hasattr(accordion, "id")
        assert accordion.pages == {}

    def test_mandatory_and_optional(self, pages_as_dict):
        accordion = vm.Accordion(id="accordion", pages=pages_as_dict)
        assert hasattr(accordion, "id")
        assert accordion.pages == pages_as_dict

    def test_valid_pages_as_list(self, pages_as_list):
        accordion = vm.Accordion(pages=pages_as_list)
        assert accordion.pages == {ACCORDION_DEFAULT_TITLE: pages_as_list}

    @pytest.mark.parametrize("pages", [{"Group": []}, []])
    def test_invalid_field_pages_no_ids_provided(self, pages):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            vm.Accordion(pages=pages)

    def test_invalid_field_pages_wrong_input_type(self):
        with pytest.raises(ValidationError, match="str type expected"):
            vm.Accordion(pages=[vm.Page(title="Page 3", components=[vm.Button()])])

    @pytest.mark.parametrize("pages", [["non existent page"], {"Group": ["non existent page"]}])
    def test_invalid_page(self, pages):
        with pytest.raises(
            ValidationError, match=re.escape("Unknown page ID ['non existent page'] provided to argument 'pages'.")
        ):
            vm.Accordion(pages=pages)


class TestAccordionBuild:
    """Tests accordion build method."""

    common_args = {"always_open": True, "persistence": True, "persistence_type": "session", "id": "accordion"}

    test_cases = [
        (
            {"Group": ["Page 1", "Page 2"]},
            dbc.Accordion(
                children=[
                    dbc.AccordionItem(
                        children=[
                            dbc.Button(children=["Page 1"], active=True, href="/", key="/"),
                            dbc.Button(children=["Page 2"], active=False, href="/page-2", key="/page-2"),
                        ],
                        title="GROUP",
                    )
                ],
                **common_args,
            ),
        ),
        (
            {"Group 1": ["Page 1"], "Group 2": ["Page 2"]},
            dbc.Accordion(
                children=[
                    dbc.AccordionItem(
                        children=[
                            dbc.Button(children=["Page 1"], active=True, href="/", key="/"),
                        ],
                        title="GROUP 1",
                    ),
                    dbc.AccordionItem(
                        children=[
                            dbc.Button(children=["Page 2"], active=False, href="/page-2", key="/page-2"),
                        ],
                        title="GROUP 2",
                    ),
                ],
                **common_args,
            ),
        ),
        (
            ["Page 1", "Page 2"],
            dbc.Accordion(
                children=[
                    dbc.AccordionItem(
                        children=[
                            dbc.Button(children=["Page 1"], active=True, href="/", key="/"),
                            dbc.Button(children=["Page 2"], active=False, href="/page-2", key="/page-2"),
                        ],
                        title=ACCORDION_DEFAULT_TITLE,
                    ),
                ],
                **common_args,
            ),
        ),
    ]

    @pytest.mark.parametrize("pages, expected", test_cases)
    def test_accordion(self, pages, expected):
        accordion = vm.Accordion(id="accordion", pages=pages).build(active_page_id="Page 1")
        assert_component_equal(accordion, html.Div(id="nav-panel"), keys_to_strip={"children"})
        assert_component_equal(accordion["accordion"], expected, keys_to_strip={"class_name", "className"})

    def test_accordion_one_page(self):
        accordion = vm.Accordion(pages={"Group": ["Page 1"]}).build(active_page_id="Page 1")
        assert_component_equal(accordion, html.Div(hidden=True, id="nav-panel"))



"""Tests for templates."""
import numpy as np
import pandas as pd
import plotly.express as px
import pytest
from plotly.graph_objs.layout._template import Template

import vizro._themes as themes
from vizro._themes._color_values import COLORS


@pytest.fixture
def template():
    """Fixture for template."""
    return np.random.choice([themes.dark, themes.light], 1)[0]


@pytest.fixture
def chart_data():
    """Fixture for chart data."""
    data_points = np.random.randint(0, 100, 1, dtype=int)
    data = pd.DataFrame(
        {
            "x": np.random.choice(["A", "B", "C", "D", "E", "F", "G", "H", "I", "J"], data_points),
            "y": np.random.normal(0, 200, data_points),
        }
    )
    return data


def test_template(template):
    assert isinstance(template, Template)


def test_color_palette_applied(chart_data, template):
    fig_1 = px.scatter(chart_data, x="x", y="y", color="y")  # continuous column for color
    fig_2 = px.scatter(chart_data, x="x", y="y", color="x")  # discrete column for color

    marker_colors = [fig_2.data[i].marker.color for i in range(len(fig_2.data))]
    assert fig_1.layout.coloraxis.colorscale == template.layout.colorscale.sequential
    assert marker_colors == COLORS["DISCRETE_10"][: len(marker_colors)]



from typing import List, Literal, Optional

import dash_bootstrap_components as dbc
from dash import html

try:
    from pydantic.v1 import Field
except ImportError:  # pragma: no cov
    from pydantic import Field

from vizro.models import Action, VizroBaseModel
from vizro.models._models_utils import _log_call


class Alert(VizroBaseModel):
    """Component `Alert` to provide user feedback.

    Args:
        type (Literal["alert"]): Defaults to `"alert"`.
        text (str): Text to be displayed in the alert.
        is_open (bool): Flag indicating whether alert should be open by default. Defaults to `True`.
        duration (Optional[int]): Duration in milliseconds for the alert to appear. Defaults to `None`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["alert"] = "alert"
    text: str = Field(..., description="Text to be displayed in the alert.")
    is_open: bool = Field(True, description="Flag indicating whether alert should be open by default.")
    duration: Optional[int] = Field(None, description="Duration in milliseconds for the alert to appear.", ge=0)
    actions: List[Action] = []

    @_log_call
    def build(self):
        return html.Div(
            [
                dbc.Alert(
                    id=self.id,
                    children=[html.P(self.text)],
                    duration=self.duration,
                    is_open=self.is_open,
                    className="alert",
                ),
            ],
            className="alert_container",
        )



"""Unit tests for vizro.actions._action_loop._get_action_loop_components file."""

import json

import plotly
import pytest
from dash import dcc, html

import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro
from vizro.actions import export_data, filter_interaction
from vizro.actions._action_loop._get_action_loop_components import _get_action_loop_components
from vizro.managers import model_manager


@pytest.fixture
def fundamental_components():
    return [
        dcc.Store(id="action_finished"),
        dcc.Store(id="remaining_actions", data=[]),
        html.Div(id="cycle_breaker_div", hidden=True),
        dcc.Store(id="cycle_breaker_empty_output_store"),
    ]


@pytest.fixture
def gateway_components(request):
    components = request.param
    actions_chain_ids = [model_manager[component].actions[0].id for component in components]
    return [
        dcc.Store(
            id={"type": "gateway_input", "trigger_id": actions_chain_id},
            data=f"{actions_chain_id}",
        )
        for actions_chain_id in actions_chain_ids
    ]


@pytest.fixture
def action_trigger_components(request):
    components = request.param
    actions_ids = [model_manager[component].actions[0].actions[0].id for component in components]
    return [dcc.Store(id={"type": "action_trigger", "action_name": action_id}) for action_id in actions_ids]


@pytest.fixture
def action_trigger_actions_id_component(request):
    components = request.param
    actions_ids = [model_manager[component].actions[0].actions[0].id for component in components]
    return dcc.Store(
        id="action_trigger_actions_id",
        data=actions_ids,
    )


@pytest.fixture
def trigger_to_actions_chain_mapper_component(request):
    components = request.param
    actions_chain_ids = [model_manager[component].actions[0].id for component in components]
    return dcc.Store(
        id="trigger_to_actions_chain_mapper",
        data={
            actions_chain_id: [action.id for action in model_manager[actions_chain_id].actions]
            for actions_chain_id in actions_chain_ids
        },
    )


@pytest.fixture
def managers_one_page_two_components_two_controls(vizro_app, dash_data_table_with_id):
    """Instantiates managers with one page that contains two controls and two components."""
    vm.Dashboard(
        pages=[
            vm.Page(
                id="test_page",
                title="First page",
                components=[
                    vm.Table(
                        id="vizro_table",
                        figure=dash_data_table_with_id,
                        actions=[
                            vm.Action(
                                id="table_filter_interaction_action",
                                function=filter_interaction(targets=["scatter_chart"]),
                            )
                        ],
                    ),
                    vm.Graph(
                        id="scatter_chart",
                        figure=px.scatter(px.data.gapminder(), x="lifeExp", y="gdpPercap"),
                    ),
                    vm.Button(
                        id="export_data_button",
                        actions=[vm.Action(id="export_data_action", function=export_data())],
                    ),
                ],
                controls=[
                    vm.Filter(
                        id="filter_continent", column="continent", selector=vm.Dropdown(id="filter_continent_selector")
                    ),
                    vm.Parameter(
                        id="parameter_x",
                        targets=["scatter_chart.x"],
                        selector=vm.Dropdown(
                            id="parameter_x_selector",
                            options=["lifeExp", "gdpPercap", "pop"],
                        ),
                    ),
                ],
            )
        ]
    )

    Vizro._pre_build()


@pytest.fixture
def managers_one_page_no_actions(vizro_app):
    """Instantiates managers with one "empty" page."""
    vm.Dashboard(
        pages=[
            vm.Page(
                id="test_page_no_actions",
                title="Second page",
                components=[vm.Card(text="")],
            )
        ]
    )

    Vizro._pre_build()


class TestGetActionLoopComponents:
    """Tests getting required components for the action loop."""

    @pytest.mark.usefixtures("managers_one_page_no_actions")
    def test_no_components(self):
        result = _get_action_loop_components()
        result = json.loads(json.dumps(result, cls=plotly.utils.PlotlyJSONEncoder))

        expected = json.loads(json.dumps(html.Div(id="action_loop_components_div"), cls=plotly.utils.PlotlyJSONEncoder))

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_components_two_controls")
    @pytest.mark.parametrize(
        "gateway_components, "
        "action_trigger_components, "
        "action_trigger_actions_id_component, "
        "trigger_to_actions_chain_mapper_component",
        [
            (
                ["test_page", "vizro_table", "export_data_button", "filter_continent_selector", "parameter_x_selector"],
                ["test_page", "vizro_table", "export_data_button", "filter_continent_selector", "parameter_x_selector"],
                ["test_page", "vizro_table", "export_data_button", "filter_continent_selector", "parameter_x_selector"],
                ["test_page", "vizro_table", "export_data_button", "filter_continent_selector", "parameter_x_selector"],
            )
        ],
        indirect=True,
    )
    def test_all_action_loop_components(  # pylint: disable=too-many-arguments
        self,
        fundamental_components,
        gateway_components,
        action_trigger_components,
        action_trigger_actions_id_component,
        trigger_to_actions_chain_mapper_component,
    ):
        result = json.loads(json.dumps(_get_action_loop_components(), cls=plotly.utils.PlotlyJSONEncoder))

        all_action_loop_components_expected = (
            fundamental_components
            + gateway_components
            + action_trigger_components
            + [action_trigger_actions_id_component]
            + [trigger_to_actions_chain_mapper_component]
        )

        expected = json.loads(
            json.dumps(
                html.Div(children=all_action_loop_components_expected, id="action_loop_components_div"),
                cls=plotly.utils.PlotlyJSONEncoder,
            )
        )

        assert result == expected



import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
from vizro import Vizro
from vizro.actions import export_data, filter_interaction
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def target_scatter_filter_and_filter_interaction(request, gapminder_2007):
    pop_filter, continent_filter_interaction, country_table_filter_interaction = request.param
    data = gapminder_2007
    if pop_filter:
        data = data[data["pop"].between(pop_filter[0], pop_filter[1], inclusive="both")]
    if continent_filter_interaction:
        data = data[data["continent"].isin(continent_filter_interaction)]
    if country_table_filter_interaction:
        data = data[data["country"].isin(country_table_filter_interaction)]
    return data


@pytest.fixture
def target_box_filtered_pop(request, gapminder_2007):
    pop_filter = request.param
    data = gapminder_2007
    if pop_filter:
        data = data[data["pop"].between(pop_filter[0], pop_filter[1], inclusive="both")]
    return data


@pytest.fixture
def managers_one_page_without_graphs_one_button():
    """Instantiates a simple model_manager and data_manager with a page, and no graphs."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[vm.Button(id="button")],
    )
    Vizro._pre_build()


@pytest.fixture
def callback_context_export_data(request):
    """Mock dash.callback_context that represents filters and filter interactions applied."""
    targets, pop_filter, continent_filter_interaction, country_table_filter_interaction = request.param
    args_grouping_filter_interaction = []
    if continent_filter_interaction:
        args_grouping_filter_interaction.append(
            {
                "clickData": CallbackTriggerDict(
                    id="box_chart",
                    property="clickData",
                    value={
                        "points": [
                            {
                                "customdata": [continent_filter_interaction],
                            }
                        ]
                    },
                    str_id="box_chart",
                    triggered=False,
                )
            }
        )
    if country_table_filter_interaction:
        args_grouping_filter_interaction.append(
            {
                "active_cell": CallbackTriggerDict(
                    id="underlying_table_id",
                    property="active_cell",
                    value={"row": 0, "column": 0, "column_id": "country"},
                    str_id="underlying_table_id",
                    triggered=False,
                ),
                "derived_viewport_data": CallbackTriggerDict(
                    id="underlying_table_id",
                    property="derived_viewport_data",
                    value=[
                        {
                            "country": "Algeria",
                            "continent": "Africa",
                            "year": 2007,
                        },
                        {
                            "country": "Egypt",
                            "continent": "Africa",
                            "year": 2007,
                        },
                    ],
                    str_id="underlying_table_id",
                    triggered=False,
                ),
            }
        )
    mock_callback_context = {
        "args_grouping": {
            "external": {
                "filters": [
                    CallbackTriggerDict(
                        id="pop_filter",
                        property="value",
                        value=pop_filter,
                        str_id="pop_filter",
                        triggered=False,
                    )
                ]
                if pop_filter
                else [],
                "filter_interaction": args_grouping_filter_interaction,
            },
        },
        "outputs_list": [
            {
                "id": {"action_id": "test_action", "target_id": target, "type": "download_dataframe"},
                "property": "data",
            }
            for target in targets
        ],
    }

    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


class TestExportData:
    @pytest.mark.usefixtures("managers_one_page_without_graphs_one_button")
    @pytest.mark.parametrize("callback_context_export_data", [([[], None, None, None])], indirect=True)
    def test_no_graphs_no_targets(self, callback_context_export_data):
        # Add action to relevant component
        model_manager["button"].actions = [vm.Action(id="test_action", function=export_data())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {}

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data", [([["scatter_chart", "box_chart"], None, None, None])], indirect=True
    )
    def test_graphs_no_targets(self, callback_context_export_data, gapminder_2007):
        # Add action to relevant component
        model_manager["button"].actions = [vm.Action(id="test_action", function=export_data())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "download_dataframe_scatter_chart": {
                "filename": "scatter_chart.csv",
                "content": gapminder_2007.to_csv(index=False),
                "type": None,
                "base64": False,
            },
            "download_dataframe_box_chart": {
                "filename": "box_chart.csv",
                "content": gapminder_2007.to_csv(index=False),
                "type": None,
                "base64": False,
            },
        }

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data, targets",
        [
            ([["scatter_chart", "box_chart"], None, None, None], None),
            ([["scatter_chart", "box_chart"], None, None, None], []),
        ],
        indirect=["callback_context_export_data"],
    )
    def test_graphs_false_targets(self, callback_context_export_data, targets, gapminder_2007):
        # Add action to relevant component
        model_manager["button"].actions = [vm.Action(id="test_action", function=export_data(targets=targets))]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "download_dataframe_scatter_chart": {
                "filename": "scatter_chart.csv",
                "content": gapminder_2007.to_csv(index=False),
                "type": None,
                "base64": False,
            },
            "download_dataframe_box_chart": {
                "filename": "box_chart.csv",
                "content": gapminder_2007.to_csv(index=False),
                "type": None,
                "base64": False,
            },
        }

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize("callback_context_export_data", [(["scatter_chart"], None, None, None)], indirect=True)
    def test_one_target(self, callback_context_export_data, gapminder_2007):
        # Add action to relevant component
        model_manager["button"].actions = [vm.Action(id="test_action", function=export_data(targets=["scatter_chart"]))]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "download_dataframe_scatter_chart": {
                "filename": "scatter_chart.csv",
                "content": gapminder_2007.to_csv(index=False),
                "type": None,
                "base64": False,
            },
        }

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data", [(["scatter_chart", "box_chart"], None, None, None)], indirect=True
    )
    def test_multiple_targets(self, callback_context_export_data, gapminder_2007):
        # Add action to relevant component
        model_manager["button"].actions = [
            vm.Action(id="test_action", function=export_data(targets=["scatter_chart", "box_chart"]))
        ]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "download_dataframe_scatter_chart": {
                "filename": "scatter_chart.csv",
                "content": gapminder_2007.to_csv(index=False),
                "type": None,
                "base64": False,
            },
            "download_dataframe_box_chart": {
                "filename": "box_chart.csv",
                "content": gapminder_2007.to_csv(index=False),
                "type": None,
                "base64": False,
            },
        }

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize("callback_context_export_data", [(["invalid_target_id"], None, None, None)], indirect=True)
    def test_invalid_target(
        self,
        callback_context_export_data,
    ):
        # Add action to relevant component
        model_manager["button"].actions = [
            vm.Action(id="test_action", function=export_data(targets=["invalid_target_id"]))
        ]

        with pytest.raises(ValueError, match="Component 'invalid_target_id' does not exist."):
            # Run action by picking the above added action function and executing it with ()
            model_manager["test_action"].function()

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data, target_scatter_filter_and_filter_interaction, target_box_filtered_pop",
        [
            (
                [["scatter_chart", "box_chart"], [10**6, 10**7], None, None],
                [[10**6, 10**7], None, None],
                [10**6, 10**7],
            ),
            ([["scatter_chart", "box_chart"], None, "Africa", None], [None, ["Africa"], None], None),
            (
                [["scatter_chart", "box_chart"], [10**6, 10**7], "Africa", None],
                [[10**6, 10**7], ["Africa"], None],
                [10**6, 10**7],
            ),
        ],
        indirect=True,
    )
    def test_multiple_targets_with_filter_and_filter_interaction(
        self,
        callback_context_export_data,
        target_scatter_filter_and_filter_interaction,
        target_box_filtered_pop,
    ):
        # Creating and adding a Filter object to the existing Page
        pop_filter = vm.Filter(column="pop", selector=vm.RangeSlider(id="pop_filter"))
        model_manager["test_page"].controls = [pop_filter]
        # Adds a default _filter Action to the filter selector objects
        pop_filter.pre_build()

        # Add filter_interaction Action to scatter_chart component
        model_manager["box_chart"].actions = [
            vm.Action(id="filter_interaction", function=filter_interaction(targets=["scatter_chart"]))
        ]

        # Add export_data action to relevant component
        model_manager["button"].actions = [
            vm.Action(id="test_action", function=export_data(targets=["scatter_chart", "box_chart"]))
        ]

        # Run action by picking the above added export_data action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "download_dataframe_scatter_chart": {
                "filename": "scatter_chart.csv",
                "content": target_scatter_filter_and_filter_interaction.to_csv(index=False),
                "type": None,
                "base64": False,
            },
            "download_dataframe_box_chart": {
                "filename": "box_chart.csv",
                "content": target_box_filtered_pop.to_csv(index=False),
                "type": None,
                "base64": False,
            },
        }

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_table_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data, target_scatter_filter_and_filter_interaction, target_box_filtered_pop",
        [
            (
                [["scatter_chart", "box_chart"], [10**6, 10**7], None, "Algeria"],
                [[10**6, 10**7], None, ["Algeria"]],
                [10**6, 10**7],
            ),
            ([["scatter_chart", "box_chart"], None, "Africa", "Algeria"], [None, ["Africa"], ["Algeria"]], None),
            (
                [["scatter_chart", "box_chart"], [10**6, 10**7], "Africa", "Algeria"],
                [[10**6, 10**7], ["Africa"], ["Algeria"]],
                [10**6, 10**7],
            ),
        ],
        indirect=True,
    )
    def test_multiple_targets_with_filter_and_filter_interaction_and_table(
        self,
        callback_context_export_data,
        target_scatter_filter_and_filter_interaction,
        target_box_filtered_pop,
    ):
        # Creating and adding a Filter object to the existing Page
        pop_filter = vm.Filter(column="pop", selector=vm.RangeSlider(id="pop_filter"))
        model_manager["test_page"].controls = [pop_filter]
        # Adds a default _filter Action to the filter selector objects
        pop_filter.pre_build()

        # Add filter_interaction Action to scatter_chart component
        model_manager["box_chart"].actions = [
            vm.Action(id="filter_interaction", function=filter_interaction(targets=["scatter_chart"]))
        ]

        # Add table filter_interaction Action to scatter_chart component
        model_manager["vizro_table"].actions = [vm.Action(function=filter_interaction(targets=["scatter_chart"]))]
        model_manager["vizro_table"].pre_build()

        # Add export_data action to relevant component
        model_manager["button"].actions = [
            vm.Action(id="test_action", function=export_data(targets=["scatter_chart", "box_chart"]))
        ]

        # Run action by picking the above added export_data action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "download_dataframe_scatter_chart": {
                "filename": "scatter_chart.csv",
                "content": target_scatter_filter_and_filter_interaction.to_csv(index=False),
                "type": None,
                "base64": False,
            },
            "download_dataframe_box_chart": {
                "filename": "box_chart.csv",
                "content": target_box_filtered_pop.to_csv(index=False),
                "type": None,
                "base64": False,
            },
        }

        assert result == expected



import json
from pathlib import Path

import dash
import dash_bootstrap_components as dbc
import plotly
import pytest
from asserts import assert_component_equal
from dash import html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro
import vizro.models as vm
from vizro import Vizro
from vizro.actions._action_loop._action_loop import ActionLoop
from vizro.models._dashboard import _all_hidden


class TestDashboardInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_dashboard_mandatory_only(self, page_1, page_2):
        dashboard = vm.Dashboard(pages=[page_1, page_2])
        assert hasattr(dashboard, "id")
        assert dashboard.pages == [page_1, page_2]
        assert dashboard.theme == "vizro_dark"
        assert dashboard.title == ""
        assert isinstance(dashboard.navigation, vm.Navigation)
        assert dashboard.navigation.pages == ["Page 1", "Page 2"]

    def test_create_dashboard_mandatory_and_optional(self, page_1, page_2):
        dashboard = vm.Dashboard(pages=[page_1, page_2], theme="vizro_light", title="Vizro")
        assert hasattr(dashboard, "id")
        assert dashboard.pages == [page_1, page_2]
        assert dashboard.theme == "vizro_light"
        assert dashboard.title == "Vizro"
        assert isinstance(dashboard.navigation, vm.Navigation)
        assert dashboard.navigation.pages == ["Page 1", "Page 2"]

    def test_navigation_pages_automatically_populated(self, page_1, page_2):
        dashboard = vm.Dashboard(pages=[page_1, page_2])
        assert dashboard.navigation.pages == ["Page 1", "Page 2"]

    def test_navigation_with_pages(self, page_1, page_2):
        dashboard = vm.Dashboard(pages=[page_1, page_2], navigation=vm.Navigation(pages=["Page 1"]))
        assert dashboard.navigation.pages == ["Page 1"]

    def test_mandatory_pages_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Dashboard()

    def test_field_invalid_pages_empty_list(self):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            vm.Dashboard(pages=[])

    def test_field_invalid_pages_input_type(self):
        with pytest.raises(ValidationError, match="4 validation errors for Dashboard"):
            vm.Dashboard(pages=[vm.Button()])

    def test_field_invalid_theme_input_type(self, page_1):
        with pytest.raises(ValidationError, match="unexpected value; permitted: 'vizro_dark', 'vizro_light'"):
            vm.Dashboard(pages=[page_1], theme="not_existing")


class TestDashboardPreBuild:
    """Tests dashboard pre_build method."""

    def test_page_registry(self, vizro_app, page_1, page_2, mocker):
        mock_register_page = mocker.patch("dash.register_page", autospec=True)
        mock_make_page_404_layout = mocker.patch(
            "vizro.models._dashboard.Dashboard._make_page_404_layout"
        )  # Checking the actual dash components is done in test_make_page_404_layout.
        vm.Dashboard(pages=[page_1, page_2]).pre_build()

        mock_register_page.assert_any_call(
            module=page_1.id,
            name="Page 1",
            description="",
            image=None,
            title="Page 1",
            path="/",
            order=0,
            layout=mocker.ANY,  # partial call is tricky to mock out so we ignore it.
        )
        mock_register_page.assert_any_call(
            module=page_2.id,
            name="Page 2",
            description="",
            image=None,
            title="Page 2",
            path="/page-2",
            order=1,
            layout=mocker.ANY,  # partial call is tricky to mock out so we ignore it.
        )
        mock_register_page.assert_any_call(
            module="not_found_404",
            layout=mock_make_page_404_layout(),
        )
        assert mock_register_page.call_count == 3

    def test_page_registry_with_title(self, vizro_app, page_1, mocker):
        mock_register_page = mocker.patch("dash.register_page", autospec=True)
        vm.Dashboard(pages=[page_1], title="My dashboard").pre_build()

        mock_register_page.assert_any_call(
            module=page_1.id,
            name="Page 1",
            description="",
            image=None,
            title="My dashboard: Page 1",
            path="/",
            order=0,
            layout=mocker.ANY,  # partial call is tricky to mock out so we ignore it.
        )

    def test_page_registry_with_description(self, vizro_app, mocker):
        mock_register_page = mocker.patch("dash.register_page", autospec=True)
        vm.Dashboard(
            pages=[vm.Page(title="Page 1", components=[vm.Button()], description="My description")]
        ).pre_build()

        mock_register_page.assert_any_call(
            module="Page 1",
            name="Page 1",
            description="My description",
            image=None,
            title="Page 1",
            path="/",
            order=0,
            layout=mocker.ANY,  # partial call is tricky to mock out so we ignore it.
        )

    @pytest.mark.parametrize(
        "image_path", ["app.png", "app.svg", "images/app.png", "images/app.svg", "logo.png", "logo.svg"]
    )
    def test_page_registry_with_image(self, page_1, mocker, tmp_path, image_path):
        if Path(image_path).parent != Path("."):
            Path(tmp_path / image_path).parent.mkdir()
        Path(tmp_path / image_path).touch()
        Vizro(assets_folder=tmp_path)
        mock_register_page = mocker.patch("dash.register_page", autospec=True)
        vm.Dashboard(pages=[page_1]).pre_build()

        mock_register_page.assert_any_call(
            module=page_1.id,
            name="Page 1",
            description="",
            image=image_path,
            title="Page 1",
            path="/",
            order=0,
            layout=mocker.ANY,  # partial call is tricky to mock out so we ignore it.
        )

    def test_page_registry_with_images(self, page_1, mocker, tmp_path):
        Path(tmp_path / "app.svg").touch()
        Path(tmp_path / "logo.svg").touch()
        Vizro(assets_folder=tmp_path)
        mock_register_page = mocker.patch("dash.register_page", autospec=True)
        vm.Dashboard(pages=[page_1]).pre_build()

        mock_register_page.assert_any_call(
            module=page_1.id,
            name="Page 1",
            description="",
            image="app.svg",
            title="Page 1",
            path="/",
            order=0,
            layout=mocker.ANY,  # partial call is tricky to mock out so we ignore it.
        )

    def test_make_page_404_layout(self, vizro_app):
        # vizro_app fixture is needed to avoid mocking out get_relative_path.
        expected = html.Div(
            [
                html.Img(src="/vizro/images/errors/error_404.svg"),
                html.Div(
                    [
                        html.Div(
                            [
                                html.H3("This page could not be found.", className="heading-3-600"),
                                html.P("Make sure the URL you entered is correct."),
                            ],
                            className="error_text_container",
                        ),
                        dbc.Button("Take me home", href="/", className="button_primary"),
                    ],
                    className="error_content_container",
                ),
            ],
            className="page_error_container",
        )

        assert_component_equal(vm.Dashboard._make_page_404_layout(), expected)


class TestDashboardBuild:
    """Tests dashboard build method."""

    def test_dashboard_build(self, vizro_app, page_1, page_2):
        dashboard = vm.Dashboard(pages=[page_1, page_2])
        dashboard.pre_build()

        dashboard_container = html.Div(
            id="dashboard_container_outer",
            children=[
                html.Div(vizro.__version__, id="vizro_version", hidden=True),
                ActionLoop._create_app_callbacks(),
                dash.page_container,
            ],
            className="vizro_dark",
        )
        result = json.loads(json.dumps(dashboard.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(dashboard_container, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected


@pytest.mark.parametrize(
    "components, expected",
    [
        ([html.Div(hidden=False)], False),
        ([html.Div(id="children-id")], False),
        ([html.Div(hidden=True), None, html.Div(hidden=False)], False),
        ([html.Div(hidden=True), None, html.Div(id="children-id")], False),
        ([html.Div(hidden=True)], True),
        ([None], True),
        ([html.Div(hidden=True), None], True),
    ],
)
def test_get_hideable_parent_div_visible(components, expected):
    assert _all_hidden(components) == expected



import re

import pandas as pd
import pytest

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
import vizro.plotly.express as px
from vizro._constants import ON_PAGE_LOAD_ACTION_PREFIX
from vizro.models._action._actions_chain import ActionsChain


class TestPageInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_page_mandatory_only(self):
        page = vm.Page(title="Page 1", components=[vm.Button(), vm.Button()])
        assert isinstance(page.components[0], vm.Button) and isinstance(page.components[1], vm.Button)
        assert page.layout.grid == [[0], [1]]
        assert page.controls == []
        assert page.title == "Page 1"
        assert page.id == "Page 1"
        assert page.path == "/page-1"
        assert page.actions == []

    def test_create_page_mandatory_and_optional(self):
        page = vm.Page(
            title="Page 1",
            components=[vm.Button(), vm.Button()],
            id="my-id",
            layout=vm.Layout(grid=[[0, 1]]),
            path="my-path",
        )
        assert isinstance(page.components[0], vm.Button) and isinstance(page.components[1], vm.Button)
        assert page.layout.grid == [[0, 1]]
        assert page.controls == []
        assert page.title == "Page 1"
        assert page.id == "my-id"
        assert page.path == "/my-path"
        assert page.actions == []

    def test_mandatory_title_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Page(id="my-id", components=[vm.Button()])

    def test_mandatory_components_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Page(title="Page 1")

    def test_mandatory_components_invalid(self):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            vm.Page(title="Page 5", components=[])

    def test_set_id_duplicate_title_valid(self):
        vm.Page(title="Page 1", components=[vm.Button()], id="my-id-1")
        vm.Page(title="Page 1", components=[vm.Button()], id="my-id-2")

    def test_set_id_duplicate_title_invalid(self):
        with pytest.raises(
            ValueError,
            match="Page with id=Page 1 already exists. Page id is automatically set to the same as the page title. "
            "If you have multiple pages with the same title then you must assign a unique id.",
        ):
            vm.Page(title="Page 1", components=[vm.Button()])
            vm.Page(title="Page 1", components=[vm.Button()])

    @pytest.mark.parametrize(
        "test_path, expected",
        [
            ("this-path-works", "/this-path-works"),
            ("2147abc", "/2147abc"),
            ("this_path_works", "/this_path_works"),
            ("this/path/works", "/this/path/works"),
            ("", "/page-12"),
        ],
    )
    def test_set_path_valid(self, test_path, expected):
        page = vm.Page(title="Page 1/2", components=[vm.Button()], path=test_path)
        assert page.path == expected

    @pytest.mark.parametrize("test_path", ["this needs? fixing*", " this needs fixing", "THIS NEEDS FIXING"])
    def test_set_path_invalid(self, test_path):
        page = vm.Page(title="Page 1", components=[vm.Button()], path=test_path)
        assert page.path == "/this-needs-fixing"

    def test_set_layout_valid(self):
        vm.Page(title="Page 1", components=[vm.Button(), vm.Button()], layout=vm.Layout(grid=[[0, 1]]))

    def test_set_layout_invalid(self):
        with pytest.raises(ValidationError, match="Number of page and grid components need to be the same."):
            vm.Page(title="Page 4", components=[vm.Button()], layout=vm.Layout(grid=[[0, 1]]))

    def test_valid_component_types(self, standard_px_chart, standard_dash_table):
        vm.Page(
            title="Page Title",
            components=[
                vm.Graph(figure=standard_px_chart),
                vm.Card(text="""# Header 1"""),
                vm.Button(),
                vm.Table(figure=standard_dash_table),
            ],
        )

    @pytest.mark.parametrize(
        "test_component",
        [vm.Checklist(), vm.Dropdown(), vm.RadioItems(), vm.RangeSlider(), vm.Slider()],
    )
    def test_invalid_component_types(self, test_component):
        with pytest.raises(ValidationError, match=re.escape("(allowed values: 'button', 'card', 'graph', 'table')")):
            vm.Page(title="Page Title", components=[test_component])

    def test_valid_control_types(self, standard_px_chart):
        vm.Page(
            title="Page Title",
            components=[vm.Graph(id="scatter", figure=standard_px_chart)],
            controls=[
                vm.Filter(column="continent"),
                vm.Parameter(targets=["scatter.x"], selector=vm.RadioItems(options=["lifeExp", "pop", "gdpPercap"])),
            ],
        )

    @pytest.mark.parametrize(
        "test_control",
        [
            vm.Button(),
            vm.Card(text="""Text"""),
            vm.Graph(figure=px.bar(data_frame=pd.DataFrame({"x": ["A", "B"], "y": [1, 2]}), x="x", y="y")),
            vm.Checklist(),
            vm.Dropdown(),
            vm.RadioItems(),
            vm.RangeSlider(),
            vm.Slider(),
        ],
    )
    def test_invalid_control_types(self, test_control):
        with pytest.raises(ValidationError, match=re.escape("(allowed values: 'filter', 'parameter')")):
            vm.Page(title="Page Title", components=[vm.Button()], controls=[test_control])


# TODO: Remove this if we can get rid of on-page-load action
class TestPagePreBuildMethod:
    def test_action_auto_generation_valid(self, standard_px_chart):
        page = vm.Page(title="Page 1", components=[vm.Graph(id="scatter_chart", figure=standard_px_chart)])
        page.pre_build()
        assert len(page.actions) == 1
        assert isinstance(page.actions[0], ActionsChain)
        assert page.actions[0].id == f"{ON_PAGE_LOAD_ACTION_PREFIX}_Page 1"


# TODO: Add unit tests for private methods in page build



"""Dark themed plotly template."""
from plotly import graph_objects as go
from plotly.graph_objs.layout._template import Template

from vizro._themes._color_values import COLORS
from vizro._themes._templates.common_values import create_template_common


def create_template_dark() -> Template:
    """Create dark themed plotly template.

    Returns:
        A plotly template object containing the dark theme
    """
    template_dark = create_template_common()

    # LAYOUT
    template_dark["layout"]["font"]["color"] = COLORS["WHITE_85"]
    template_dark["layout"]["title"]["font"]["color"] = COLORS["WHITE_85"]
    template_dark["layout"]["legend"]["font"]["color"] = COLORS["WHITE_85"]
    template_dark["layout"]["legend"]["title"]["font"]["color"] = COLORS["WHITE_85"]
    template_dark["layout"]["paper_bgcolor"] = COLORS["DARK_BG03"]
    template_dark["layout"]["plot_bgcolor"] = COLORS["DARK_BG03"]
    template_dark["layout"]["geo"]["bgcolor"] = COLORS["DARK_BG03"]
    template_dark["layout"]["geo"]["lakecolor"] = COLORS["DARK_BG03"]
    template_dark["layout"]["geo"]["landcolor"] = COLORS["DARK_BG03"]
    template_dark["layout"]["polar"]["bgcolor"] = COLORS["DARK_BG03"]
    template_dark["layout"]["polar"]["angularaxis"]["gridcolor"] = COLORS["WHITE_12"]
    template_dark["layout"]["polar"]["angularaxis"]["linecolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["polar"]["radialaxis"]["gridcolor"] = COLORS["WHITE_12"]
    template_dark["layout"]["polar"]["radialaxis"]["linecolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["ternary"]["bgcolor"] = COLORS["DARK_BG03"]
    template_dark["layout"]["ternary"]["aaxis"]["gridcolor"] = COLORS["WHITE_12"]
    template_dark["layout"]["ternary"]["aaxis"]["linecolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["ternary"]["baxis"]["gridcolor"] = COLORS["WHITE_12"]
    template_dark["layout"]["ternary"]["baxis"]["linecolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["ternary"]["caxis"]["gridcolor"] = COLORS["WHITE_12"]
    template_dark["layout"]["ternary"]["caxis"]["linecolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["mapbox"]["style"] = "carto-darkmatter"
    template_dark["layout"]["coloraxis"]["colorbar"]["tickcolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["coloraxis"]["colorbar"]["tickfont"]["color"] = COLORS["WHITE_55"]
    template_dark["layout"]["coloraxis"]["colorbar"]["title"]["font"]["color"] = COLORS["WHITE_55"]
    template_dark["layout"]["colorscale"]["diverging"] = COLORS["DIVERGING_RED_CYAN"][::-1]
    template_dark["layout"]["colorscale"]["sequential"] = COLORS["DIVERGING_RED_CYAN"]  # default for continuous
    template_dark["layout"]["colorscale"]["sequentialminus"] = COLORS["SEQUENTIAL_RED"][::-1]
    template_dark["layout"]["colorway"] = COLORS["DISCRETE_10"]

    # X AXIS
    template_dark["layout"]["xaxis"]["title"]["font"]["color"] = COLORS["WHITE_85"]
    template_dark["layout"]["xaxis"]["tickcolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["xaxis"]["tickfont"]["color"] = COLORS["WHITE_55"]
    template_dark["layout"]["xaxis"]["linecolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["xaxis"]["gridcolor"] = COLORS["WHITE_12"]

    # Y AXIS
    template_dark["layout"]["yaxis"]["title"]["font"]["color"] = COLORS["WHITE_85"]
    template_dark["layout"]["yaxis"]["tickcolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["yaxis"]["tickfont"]["color"] = COLORS["WHITE_55"]
    template_dark["layout"]["yaxis"]["linecolor"] = COLORS["WHITE_30"]
    template_dark["layout"]["yaxis"]["gridcolor"] = COLORS["WHITE_12"]

    # ANNOTATIONS
    template_dark["layout"]["annotationdefaults"] = {
        "font": {"color": COLORS["WHITE_85"], "size": 14},
        "showarrow": False,
    }

    # CHART TYPES
    template_dark.data.bar = [
        go.Bar(
            marker={"line": {"color": template_dark["layout"]["paper_bgcolor"], "width": 1}},
            textfont={"color": COLORS["WHITE_55"]},
        )
    ]

    template_dark.data.table = [
        go.Table(
            header={
                "fill_color": COLORS["DARK_BG03"],
                "line_color": COLORS["WHITE_12"],
                "height": 32,
                "font": {"color": COLORS["WHITE_85"], "size": 14},
                "align": "left",
            },
            cells={
                "line_color": COLORS["WHITE_12"],
                "fill_color": COLORS["DARK_BG03"],
                "height": 32,
                "font": {"color": COLORS["WHITE_55"], "size": 14},
                "align": "left",
            },
        )
    ]

    template_dark.data.waterfall = [
        go.Waterfall(
            decreasing={"marker": {"color": COLORS["DISCRETE_10"][1]}},
            increasing={"marker": {"color": COLORS["DISCRETE_10"][0]}},
            totals={"marker": {"color": COLORS["GREY_55"]}},
            textfont_color=template_dark.layout.title.font.color,
            textposition="outside",
            connector={
                "line": {
                    "color": template_dark.layout.xaxis.tickcolor,
                    "width": 1,
                }
            },
        )
    ]
    return template_dark


dark = create_template_dark()



"""Unit tests for vizro.models.Button."""

import dash_bootstrap_components as dbc
import pytest
from asserts import assert_component_equal
from dash import html

import vizro.models as vm
from vizro.actions import export_data


class TestButtonInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_default_button(self):
        button = vm.Button()
        assert hasattr(button, "id")
        assert button.type == "button"
        assert button.text == "Click me!"
        assert button.actions == []

    @pytest.mark.parametrize(
        "text",
        ["Test", 123, 1.23, True, """# Header""", """<p>Hello </p>"""],
    )
    def test_create_button_with_optional(self, text):
        button = vm.Button(text=text)
        assert hasattr(button, "id")
        assert button.type == "button"
        assert button.text == str(text)
        assert button.actions == []

    def test_set_action_via_validator(self):
        button = vm.Button(actions=[vm.Action(function=export_data())])
        actions_chain = button.actions[0]
        assert actions_chain.trigger.component_property == "n_clicks"


class TestBuildMethod:
    def test_button_build(self):
        button = vm.Button(id="button_id", text="My text").build()
        expected = html.Div(
            dbc.Button(id="button_id", children="My text", className="button_primary"),
            className="button_container",
            id="button_id_outer",
        )
        assert_component_equal(button, expected)



"""Unit tests for vizro.models.Dropdown."""
import json

import plotly
import pytest
from dash import dcc, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

from vizro.models._action._action import Action
from vizro.models._components.form import Dropdown


@pytest.fixture()
def expected_dropdown_with_all():
    return html.Div(
        [
            html.P("Title"),
            dcc.Dropdown(
                id="dropdown_id",
                options=["ALL", "A", "B", "C"],
                value="ALL",
                multi=True,
                persistence=True,
                persistence_type="session",
                className="selector_body_dropdown",
            ),
        ],
        className="selector_dropdown_container",
        id="dropdown_id_outer",
    )


@pytest.fixture()
def expected_dropdown_without_all():
    return html.Div(
        [
            html.P("Title"),
            dcc.Dropdown(
                id="dropdown_id",
                options=["A", "B", "C"],
                value="A",
                multi=False,
                persistence=True,
                persistence_type="session",
                className="selector_body_dropdown",
            ),
        ],
        className="selector_dropdown_container",
        id="dropdown_id_outer",
    )


class TestDropdownInstantiation:
    """Tests model instantiation."""

    def test_create_dropdown_mandatory_only(self):
        dropdown = Dropdown()

        assert hasattr(dropdown, "id")
        assert dropdown.type == "dropdown"
        assert dropdown.options == []
        assert dropdown.value is None
        assert dropdown.multi is True
        assert dropdown.title == ""
        assert dropdown.actions == []

    def test_create_dropdown_mandatory_and_optional(self):
        dropdown = Dropdown(options=["A", "B", "C"], value="A", multi=False, title="Title", id="dropdown-id")

        assert dropdown.id == "dropdown-id"
        assert dropdown.type == "dropdown"
        assert dropdown.options == ["A", "B", "C"]
        assert dropdown.value == "A"
        assert dropdown.multi is False
        assert dropdown.title == "Title"
        assert dropdown.actions == []

    @pytest.mark.parametrize(
        "test_options, expected",
        [
            (["A", "B", "C"], ["A", "B", "C"]),
            ([1, 2, 3], [1.0, 2.0, 3.0]),
            ([1.0, 2.0, 3.0], [1.0, 2.0, 3.0]),
            ([True, False, True], [True, False, True]),
            (
                [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}],
                [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}],
            ),
            (
                [{"label": "New York", "value": "NYC"}, {"label": "Berlin", "value": "BER"}],
                [{"label": "New York", "value": "NYC"}, {"label": "Berlin", "value": "BER"}],
            ),
            (
                [{"label": "True", "value": True}, {"label": "False", "value": False}],
                [{"label": "True", "value": True}, {"label": "False", "value": False}],
            ),
            ([True, 2.0, 1.0, "A", "B"], ["True", "2.0", "1.0", "A", "B"]),
        ],
    )
    def test_create_dropdown_valid_options(self, test_options, expected):
        dropdown = Dropdown(options=test_options)

        assert hasattr(dropdown, "id")
        assert dropdown.type == "dropdown"
        assert dropdown.options == expected
        assert dropdown.value is None
        assert dropdown.multi is True
        assert dropdown.title == ""
        assert dropdown.actions == []

    @pytest.mark.parametrize("test_options", [1, "A", True, 1.0])
    def test_create_dropdown_invalid_options_type(self, test_options):
        with pytest.raises(ValidationError, match="value is not a valid list"):
            Dropdown(options=test_options)

    def test_create_dropdown_invalid_options_dict(self):
        with pytest.raises(
            ValidationError,
            match="Invalid argument `options` passed. Expected a dict with keys `label` and `value`.",
        ):
            Dropdown(options=[{"hello": "A", "world": "A"}, {"hello": "B", "world": "B"}])

    @pytest.mark.parametrize(
        "test_value, options, multi",
        [
            # Single default value with multi=False
            ("A", ["A", "B", "C"], False),
            (1, [1, 2, 3], False),
            (1.0, [1.0, 2.0, 3.0], False),
            (False, [True, False], False),
            ("A", [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}], False),
            ("True", [True, 2.0, 1.0, "A", "B"], False),
            # Single default value with multi=True
            ("A", ["A", "B", "C"], True),
            (1, [1, 2, 3], True),
            (1.0, [1.0, 2.0, 3.0], True),
            (False, [True, False], True),
            ("A", [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}], True),
            ("True", [True, 2.0, 1.0, "A", "B"], True),
            # List of default values with multi=True
            (["A", "B"], ["A", "B", "C"], True),
            ([1, 2], [1, 2, 3], True),
            ([1.0, 2.0], [1.0, 2.0, 3.0], True),
            ([False, True], [True, False], True),
            (["A", "B"], [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}], True),
            (["True", "A"], [True, 2.0, 1.0, "A", "B"], True),
        ],
    )
    def test_create_dropdown_valid_value(self, test_value, options, multi):
        dropdown = Dropdown(options=options, value=test_value, multi=multi)

        assert hasattr(dropdown, "id")
        assert dropdown.type == "dropdown"
        assert dropdown.value == test_value
        assert dropdown.multi == multi
        assert dropdown.title == ""
        assert dropdown.actions == []

    @pytest.mark.parametrize(
        "test_value, options",
        [
            (5, ["A", "B", "C"]),
            (5, [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            ("D", ["A", "B", "C"]),
            ("D", [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            (True, ["A", "B", "C"]),
            (True, [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
        ],
    )
    def test_create_dropdown_invalid_value(self, test_value, options):
        with pytest.raises(ValidationError, match="Please provide a valid value from `options`."):
            Dropdown(value=test_value, options=options)

    def test_create_dropdown_invalid_multi(self):
        with pytest.raises(ValidationError, match="Please set multi=True if providing a list of default values."):
            Dropdown(value=[1, 2], multi=False, options=[1, 2, 3, 4, 5])

    def test_set_action_via_validator(self, identity_action_function):
        dropdown = Dropdown(actions=[Action(function=identity_action_function())])
        actions_chain = dropdown.actions[0]
        assert actions_chain.trigger.component_property == "value"


class TestDropdownBuild:
    """Tests model build method."""

    def test_dropdown_with_all_option(self, expected_dropdown_with_all):
        dropdown = Dropdown(options=["A", "B", "C"], id="dropdown_id", title="Title").build()

        result = json.loads(json.dumps(dropdown, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_dropdown_with_all, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    def test_dropdown_without_all_option(self, expected_dropdown_without_all):
        dropdown = Dropdown(options=["A", "B", "C"], multi=False, id="dropdown_id", title="Title").build()

        result = json.loads(json.dumps(dropdown, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_dropdown_without_all, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected



import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
import vizro.plotly.express as px
from vizro._constants import FILTER_ACTION_PREFIX
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def target_scatter_filtered_continent_and_pop(request, gapminder_2007, scatter_params):
    continent, pop = request.param
    data = gapminder_2007[
        gapminder_2007["continent"].isin(continent) & gapminder_2007["pop"].between(pop[0], pop[1], inclusive="both")
    ]
    return px.scatter(data, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_box_filtered_continent_and_pop(request, gapminder_2007, box_params):
    continent, pop = request.param
    data = gapminder_2007[
        gapminder_2007["continent"].isin(continent) & gapminder_2007["pop"].between(pop[0], pop[1], inclusive="both")
    ]
    return px.box(data, **box_params).update_layout(margin_t=24)


@pytest.fixture
def callback_context_filter_continent(request):
    """Mock dash.callback_context that represents continent Filter value selection."""
    continent = request.param
    mock_callback_context = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [
                    CallbackTriggerDict(
                        id="continent_filter",
                        property="value",
                        value=continent,
                        str_id="continent_filter",
                        triggered=False,
                    )
                ],
                "parameters": [],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="on",
                    value=True,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.fixture
def callback_context_filter_continent_and_pop(request):
    """Mock dash.callback_context that represents continent and pop Filter value selection."""
    continent, pop = request.param
    mock_callback_context = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [
                    CallbackTriggerDict(
                        id="continent_filter",
                        property="value",
                        value=continent,
                        str_id="continent_filter",
                        triggered=False,
                    ),
                    CallbackTriggerDict(
                        id="pop_filter",
                        property="value",
                        value=pop,
                        str_id="pop_filter",
                        triggered=False,
                    ),
                ],
                "parameters": [],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="on",
                    value=True,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
class TestFilter:
    @pytest.mark.parametrize(
        "callback_context_filter_continent,target_scatter_filtered_continent,target_box_filtered_continent",
        [(["Africa"], ["Africa"], ["Africa"]), (["Africa", "Europe"], ["Africa", "Europe"], ["Africa", "Europe"])],
        indirect=True,
    )
    def test_one_filter_no_targets(
        self,
        callback_context_filter_continent,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        # Creating and adding a Filter object to the existing Page
        continent_filter = vm.Filter(id="test_filter", column="continent", selector=vm.Dropdown(id="continent_filter"))
        model_manager["test_page"].controls = [continent_filter]

        # Adds a default _filter Action to the "continent_filter" selector object
        continent_filter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{FILTER_ACTION_PREFIX}_test_filter"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
            "box_chart": target_box_filtered_continent,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_continent,target_scatter_filtered_continent",
        [(["Africa"], ["Africa"]), (["Africa", "Europe"], ["Africa", "Europe"])],
        indirect=True,
    )
    def test_one_filter_one_target(
        self,
        callback_context_filter_continent,
        target_scatter_filtered_continent,
    ):
        # Creating and adding a Filter object to the existing Page
        continent_filter = vm.Filter(
            id="test_filter", column="continent", targets=["scatter_chart"], selector=vm.Dropdown(id="continent_filter")
        )
        model_manager["test_page"].controls = [continent_filter]

        # Adds a default _filter Action to the "continent_filter" selector object
        continent_filter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{FILTER_ACTION_PREFIX}_test_filter"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_continent,target_scatter_filtered_continent,target_box_filtered_continent",
        [(["Africa"], ["Africa"], ["Africa"]), (["Africa", "Europe"], ["Africa", "Europe"], ["Africa", "Europe"])],
        indirect=True,
    )
    def test_one_filter_multiple_targets(
        self,
        callback_context_filter_continent,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        # Creating and adding a Filter object to the existing Page
        continent_filter = vm.Filter(
            id="test_filter",
            column="continent",
            targets=["scatter_chart", "box_chart"],
            selector=vm.Dropdown(id="continent_filter"),
        )
        model_manager["test_page"].controls = [continent_filter]

        # Adds a default _filter Action to the "continent_filter" selector object
        continent_filter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{FILTER_ACTION_PREFIX}_test_filter"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
            "box_chart": target_box_filtered_continent,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_continent_and_pop,target_scatter_filtered_continent_and_pop,target_box_filtered_continent_and_pop",
        [
            ([["Africa"], [10**6, 10**7]], [["Africa"], [10**6, 10**7]], [["Africa"], [10**6, 10**7]]),
            (
                [["Africa", "Europe"], [10**6, 10**7]],
                [["Africa", "Europe"], [10**6, 10**7]],
                [["Africa", "Europe"], [10**6, 10**7]],
            ),
        ],
        indirect=True,
    )
    def test_multiple_filters_no_targets(
        self,
        callback_context_filter_continent_and_pop,
        target_scatter_filtered_continent_and_pop,
        target_box_filtered_continent_and_pop,
    ):
        # Creating and adding a Filter objects to the existing Page
        continent_filter = vm.Filter(
            id="test_filter_continent", column="continent", selector=vm.Dropdown(id="continent_filter")
        )
        pop_filter = vm.Filter(id="test_filter_pop", column="pop", selector=vm.RangeSlider(id="pop_filter"))
        model_manager["test_page"].controls = [continent_filter, pop_filter]

        # Adds a default _filter Action to the filter selector objects
        continent_filter.pre_build()
        pop_filter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{FILTER_ACTION_PREFIX}_test_filter_continent"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent_and_pop,
            "box_chart": target_box_filtered_continent_and_pop,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_continent_and_pop,target_scatter_filtered_continent_and_pop",
        [
            ([["Africa"], [10**6, 10**7]], [["Africa"], [10**6, 10**7]]),
            ([["Africa", "Europe"], [10**6, 10**7]], [["Africa", "Europe"], [10**6, 10**7]]),
        ],
        indirect=True,
    )
    def test_multiple_filters_one_target(
        self,
        callback_context_filter_continent_and_pop,
        target_scatter_filtered_continent_and_pop,
    ):
        # Creating and adding a Filter objects to the existing Page
        continent_filter = vm.Filter(
            id="test_filter_continent",
            column="continent",
            targets=["scatter_chart"],
            selector=vm.Dropdown(id="continent_filter"),
        )
        pop_filter = vm.Filter(column="pop", targets=["scatter_chart"], selector=vm.RangeSlider(id="pop_filter"))
        model_manager["test_page"].controls = [continent_filter, pop_filter]

        # Adds a default _filter Action to the filter selector objects
        continent_filter.pre_build()
        pop_filter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{FILTER_ACTION_PREFIX}_test_filter_continent"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent_and_pop,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_continent_and_pop,target_scatter_filtered_continent_and_pop,target_box_filtered_continent_and_pop",
        [
            ([["Africa"], [10**6, 10**7]], [["Africa"], [10**6, 10**7]], [["Africa"], [10**6, 10**7]]),
            (
                [["Africa", "Europe"], [10**6, 10**7]],
                [["Africa", "Europe"], [10**6, 10**7]],
                [["Africa", "Europe"], [10**6, 10**7]],
            ),
        ],
        indirect=True,
    )
    def test_multiple_filters_multiple_targets(
        self,
        callback_context_filter_continent_and_pop,
        target_scatter_filtered_continent_and_pop,
        target_box_filtered_continent_and_pop,
    ):
        # Creating and adding a Filter objects to the existing Page
        continent_filter = vm.Filter(
            id="test_filter_continent",
            column="continent",
            targets=["scatter_chart", "box_chart"],
            selector=vm.Dropdown(id="continent_filter"),
        )
        pop_filter = vm.Filter(
            id="test_filter_pop",
            column="pop",
            targets=["scatter_chart", "box_chart"],
            selector=vm.RangeSlider(id="pop_filter"),
        )
        model_manager["test_page"].controls = [continent_filter, pop_filter]

        # Adds a default _filter Action to the filter selector objects
        continent_filter.pre_build()
        pop_filter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{FILTER_ACTION_PREFIX}_test_filter_continent"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent_and_pop,
            "box_chart": target_box_filtered_continent_and_pop,
        }

        assert result == expected



from __future__ import annotations

from typing import List, TypedDict

from dash import Input, Output, Patch, callback, dcc, html

try:
    from pydantic.v1 import Field, root_validator, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, root_validator, validator

from vizro._constants import ON_PAGE_LOAD_ACTION_PREFIX
from vizro.actions import _on_page_load
from vizro.managers._model_manager import DuplicateIDError
from vizro.models import Action, Layout, VizroBaseModel
from vizro.models._action._actions_chain import ActionsChain, Trigger
from vizro.models._models_utils import _log_call, get_unique_grid_component_ids

from .types import ComponentType, ControlType

# This is just used for type checking. Ideally it would inherit from some dash.development.base_component.Component
# (e.g. html.Div) as well as TypedDict, but that's not possible, and Dash does not have typing support anyway. When
# this type is used, the object is actually still a dash.development.base_component.Component, but this makes it easier
# to see what contract the component fulfills by making the expected keys explicit.
_PageBuildType = TypedDict("_PageBuildType", {"control-panel": html.Div, "components": html.Div})


class Page(VizroBaseModel):
    """A page in [`Dashboard`][vizro.models.Dashboard] with its own URL path and place in the `Navigation`.

    Args:
        components (List[ComponentType]): See [ComponentType][vizro.models.types.ComponentType]. At least one component
            has to be provided.
        title (str): Title to be displayed.
        description (str): Description for meta tags.
        layout (Layout): Layout to place components in. Defaults to `None`.
        controls (List[ControlType]): See [ControlType][vizro.models.types.ControlType]. Defaults to `[]`.
        path (str): Path to navigate to page. Defaults to `""`.

    Raises:
        ValueError: If number of page and grid components is not the same
    """

    components: List[ComponentType]
    title: str = Field(..., description="Title to be displayed.")
    description: str = Field("", description="Description for meta tags.")
    layout: Layout = None  # type: ignore[assignment]
    controls: List[ControlType] = []
    path: str = Field("", description="Path to navigate to page.")

    # TODO: Remove default on page load action if possible
    actions: List[ActionsChain] = []

    @root_validator(pre=True)
    def set_id(cls, values):
        if "title" not in values:
            return values

        values.setdefault("id", values["title"])
        return values

    @validator("components", always=True)
    def set_components(cls, components):
        if not components:
            raise ValueError("Ensure this value has at least 1 item.")
        return components

    @validator("layout", always=True)
    def set_layout(cls, layout, values) -> Layout:
        if "components" not in values:
            return layout

        if layout is None:
            grid = [[i] for i in range(len(values["components"]))]
            return Layout(grid=grid)

        unique_grid_idx = get_unique_grid_component_ids(layout.grid)
        if len(unique_grid_idx) != len(values["components"]):
            raise ValueError("Number of page and grid components need to be the same.")
        return layout

    @validator("path", always=True)
    def set_path(cls, path, values) -> str:
        # Based on how Github generates anchor links - see:
        # https://stackoverflow.com/questions/72536973/how-are-github-markdown-anchor-links-constructed.
        def clean_path(path, allowed_characters):
            path = path.strip().lower().replace(" ", "-")
            path = "".join(character for character in path if character.isalnum() or character in allowed_characters)
            return "/" + path

        # Allow "/" in path if provided by user, otherwise turn page id into suitable URL path (not allowing "/")
        if path:
            return clean_path(path, "-_/")
        return clean_path(values["id"], "-_")

    def __init__(self, **data):
        """Adds the model instance to the model manager."""
        try:
            super().__init__(**data)
        except DuplicateIDError as exc:
            raise ValueError(
                f"Page with id={self.id} already exists. Page id is automatically set to the same "
                f"as the page title. If you have multiple pages with the same title then you must assign a unique id."
            ) from exc

    @_log_call
    def pre_build(self):
        # TODO: Remove default on page load action if possible
        if any(hasattr(component, "figure") for component in self.components):
            self.actions = [
                ActionsChain(
                    id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_{self.id}",
                    trigger=Trigger(
                        component_id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_trigger_{self.id}",
                        component_property="data",
                    ),
                    actions=[
                        Action(
                            id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_action_{self.id}", function=_on_page_load(page_id=self.id)
                        )
                    ],
                )
            ]

    @_log_call
    def build(self) -> _PageBuildType:
        self._update_graph_theme()
        controls_content = [control.build() for control in self.controls]
        control_panel = (
            html.Div(children=[*controls_content], id="control-panel")
            if controls_content
            else html.Div(hidden=True, id="control-panel")
        )
        components_content = [
            html.Div(
                component.build(),
                style={
                    "gridColumn": f"{grid_coord.col_start}/{grid_coord.col_end}",
                    "gridRow": f"{grid_coord.row_start}/{grid_coord.row_end}",
                },
            )
            for component, grid_coord in zip(self.components, self.layout.component_grid_lines)
        ]
        components_container = self._create_component_container(components_content)
        return html.Div([control_panel, components_container])

    def _update_graph_theme(self):
        # The obvious way to do this would be to alter pio.templates.default, but this changes global state and so is
        # not good.
        # Putting graphs as inputs here would be a nice way to trigger the theme change automatically so that we don't

        # need the call to _update_theme inside Graph.__call__ also, but this results in an extra callback and the graph
        # flickering.
        # The code is written to be generic and extensible so that it runs _update_theme on any component with such a
        # method defined. But at the moment this just means Graphs.
        # TODO: consider making this clientside callback and then possibly we can remove the call to _update_theme in
        #  Graph.__call__ without any flickering.
        # TODO: if we do this then we should *consider* defining the callback in Graph itself rather than at Page
        #  level. This would mean multiple callbacks on one page but if it's clientside that probably doesn't matter.

        themed_components = [component for component in self.components if hasattr(component, "_update_theme")]
        if themed_components:

            @callback(
                [Output(component.id, "figure", allow_duplicate=True) for component in themed_components],
                Input("theme_selector", "on"),
                prevent_initial_call="initial_duplicate",
            )
            def update_graph_theme(theme_selector: bool):
                return [component._update_theme(Patch(), theme_selector) for component in themed_components]

    def _create_component_container(self, components_content):
        component_container = html.Div(
            children=[
                html.Div(
                    components_content,
                    style={
                        "gridRowGap": self.layout.row_gap,
                        "gridColumnGap": self.layout.col_gap,
                        "gridTemplateColumns": f"repeat({len(self.layout.grid[0])},"
                        f"minmax({self.layout.col_min_width}, 1fr))",
                        "gridTemplateRows": f"repeat({len(self.layout.grid)},"
                        f"minmax({self.layout.row_min_height}, 1fr))",
                    },
                    className="component_container_grid",
                ),
                dcc.Store(id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_trigger_{self.id}"),
            ],
            id="components",
        )
        return component_container



"""Contains utilities for the implementation of action functions."""

from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, TypedDict, Union

import pandas as pd

from vizro._constants import ALL_OPTION, NONE_OPTION
from vizro.managers import data_manager, model_manager
from vizro.managers._model_manager import ModelID
from vizro.models.types import MultiValueType, SelectorType, SingleValueType

if TYPE_CHECKING:
    from vizro.models import Action, VizroBaseModel

ValidatedNoneValueType = Union[SingleValueType, MultiValueType, None, List[None]]


class CallbackTriggerDict(TypedDict):
    """Represent dash.callback_context.args_grouping item. Shortened as 'ctd' in the code.

    Args:
        id: The component ID. If it`s a pattern matching ID, it will be a dict.
        property: The component property used in the callback.
        value: The value of the component property at the time the callback was fired.
        str_id: For pattern matching IDs, it's the stringified dict ID without white spaces.
        triggered: A boolean indicating whether this input triggered the callback.
    """

    id: ModelID
    property: Literal["clickData", "value", "n_clicks", "active_cell", "derived_viewport_data"]
    value: Optional[Any]
    str_id: str
    triggered: bool


# Utility functions for helper functions used in pre-defined actions ----
def _get_component_actions(component) -> List[Action]:
    return (
        [action for actions_chain in component.actions for action in actions_chain.actions]
        if hasattr(component, "actions")
        else []
    )


def _apply_filters(
    data_frame: pd.DataFrame,
    ctds_filters: List[CallbackTriggerDict],
    target: str,
) -> pd.DataFrame:
    for ctd in ctds_filters:
        selector_value = ctd["value"]
        selector_value = selector_value if isinstance(selector_value, list) else [selector_value]
        selector_actions = _get_component_actions(model_manager[ctd["id"]])

        for action in selector_actions:
            if (
                action.function._function.__name__ != "_filter"
                or target not in action.function["targets"]
                or ALL_OPTION in selector_value
            ):
                continue

            _filter_function = action.function["filter_function"]
            _filter_column = action.function["filter_column"]
            _filter_value = selector_value
            data_frame = data_frame[_filter_function(data_frame[_filter_column], _filter_value)]

    return data_frame


def _apply_graph_filter_interaction(
    data_frame: pd.DataFrame, target: str, ctd_filter_interaction: Dict[str, CallbackTriggerDict]
) -> pd.DataFrame:
    ctd_click_data = ctd_filter_interaction["clickData"]
    if not ctd_click_data["value"]:
        return data_frame

    source_graph_id: ModelID = ctd_click_data["id"]
    source_graph_actions = _get_component_actions(model_manager[source_graph_id])
    try:
        custom_data_columns = model_manager[source_graph_id]["custom_data"]
    except KeyError as exc:
        raise KeyError(f"No `custom_data` argument found for source graph with id {source_graph_id}.") from exc

    customdata = ctd_click_data["value"]["points"][0]["customdata"]

    for action in source_graph_actions:
        if action.function._function.__name__ != "filter_interaction" or target not in action.function["targets"]:
            continue
        for custom_data_idx, column in enumerate(custom_data_columns):
            data_frame = data_frame[data_frame[column].isin([customdata[custom_data_idx]])]

    return data_frame


def _get_parent_vizro_model(_underlying_callable_object_id: str) -> VizroBaseModel:
    from vizro.models import VizroBaseModel

    for _, vizro_base_model in model_manager._items_with_type(VizroBaseModel):
        if (
            hasattr(vizro_base_model, "_callable_object_id")
            and vizro_base_model._callable_object_id == _underlying_callable_object_id
        ):
            return vizro_base_model
    raise KeyError(
        f"No parent Vizro model found for underlying callable object with id: {_underlying_callable_object_id}."
    )


def _apply_table_filter_interaction(
    data_frame: pd.DataFrame, target: str, ctd_filter_interaction: Dict[str, CallbackTriggerDict]
) -> pd.DataFrame:
    ctd_active_cell = ctd_filter_interaction["active_cell"]
    ctd_derived_viewport_data = ctd_filter_interaction["derived_viewport_data"]
    if not ctd_active_cell["value"] or not ctd_derived_viewport_data["value"]:
        return data_frame

    # ctd_active_cell["id"] represents the underlying table id, so we need to fetch its parent Vizro Table actions.
    source_table_actions = _get_component_actions(_get_parent_vizro_model(ctd_active_cell["id"]))

    for action in source_table_actions:
        if action.function._function.__name__ != "filter_interaction" or target not in action.function["targets"]:
            continue
        column = ctd_active_cell["value"]["column_id"]
        derived_viewport_data_row = ctd_active_cell["value"]["row"]
        clicked_data = ctd_derived_viewport_data["value"][derived_viewport_data_row][column]
        data_frame = data_frame[data_frame[column].isin([clicked_data])]

    return data_frame


def _apply_filter_interaction(
    data_frame: pd.DataFrame,
    ctds_filter_interaction: List[Dict[str, CallbackTriggerDict]],
    target: str,
) -> pd.DataFrame:
    for ctd_filter_interaction in ctds_filter_interaction:
        if "clickData" in ctd_filter_interaction:
            data_frame = _apply_graph_filter_interaction(
                data_frame=data_frame,
                target=target,
                ctd_filter_interaction=ctd_filter_interaction,
            )

        if "active_cell" in ctd_filter_interaction and "derived_viewport_data" in ctd_filter_interaction:
            data_frame = _apply_table_filter_interaction(
                data_frame=data_frame,
                target=target,
                ctd_filter_interaction=ctd_filter_interaction,
            )

    return data_frame


def _validate_selector_value_none(value: Union[SingleValueType, MultiValueType]) -> ValidatedNoneValueType:
    if value == NONE_OPTION:
        return None
    elif isinstance(value, list):
        return [i for i in value if i != NONE_OPTION] or [None]
    return value


def _create_target_arg_mapping(dot_separated_strings: List[str]) -> Dict[str, List[str]]:
    results = defaultdict(list)
    for string in dot_separated_strings:
        if "." not in string:
            raise ValueError(f"Provided string {string} must contain a '.'")
        component, arg = string.split(".", 1)
        results[component].append(arg)
    return results


def _update_nested_graph_properties(
    graph_config: Dict[str, Any], dot_separated_string: str, value: Any
) -> Dict[str, Any]:
    keys = dot_separated_string.split(".")
    current_property = graph_config

    for key in keys[:-1]:
        current_property = current_property.setdefault(key, {})

    current_property[keys[-1]] = value
    return graph_config


def _get_parametrized_config(
    targets: List[ModelID], parameters: List[CallbackTriggerDict]
) -> Dict[ModelID, Dict[str, Any]]:
    parameterized_config = {}
    for target in targets:
        # TODO - avoid calling _captured_callable. Once we have done this we can remove _arguments from
        #  CapturedCallable entirely.
        graph_config = deepcopy(model_manager[target].figure._arguments)
        if "data_frame" in graph_config:
            graph_config.pop("data_frame")

        for ctd in parameters:
            selector_value = ctd[
                "value"
            ]  # TODO: needs to be refactored so that it is independent of implementation details
            if hasattr(selector_value, "__iter__") and ALL_OPTION in selector_value:  # type: ignore[operator]
                selector: SelectorType = model_manager[ctd["id"]]
                selector_value = selector.options
            selector_value = _validate_selector_value_none(selector_value)
            selector_actions = _get_component_actions(model_manager[ctd["id"]])

            for action in selector_actions:
                action_targets = _create_target_arg_mapping(action.function["targets"])

                if action.function._function.__name__ != "_parameter" or target not in action_targets:
                    continue

                for action_targets_arg in action_targets[target]:
                    graph_config = _update_nested_graph_properties(
                        graph_config=graph_config, dot_separated_string=action_targets_arg, value=selector_value
                    )

        parameterized_config[target] = graph_config

    return parameterized_config


# Helper functions used in pre-defined actions ----
def _get_filtered_data(
    targets: List[ModelID],
    ctds_filters: List[CallbackTriggerDict],
    ctds_filter_interaction: List[Dict[str, CallbackTriggerDict]],
) -> Dict[ModelID, pd.DataFrame]:
    filtered_data = {}
    for target in targets:
        data_frame = data_manager._get_component_data(target)

        data_frame = _apply_filters(
            data_frame=data_frame,
            ctds_filters=ctds_filters,
            target=target,
        )
        data_frame = _apply_filter_interaction(
            data_frame=data_frame,
            ctds_filter_interaction=ctds_filter_interaction,
            target=target,
        )

        filtered_data[target] = data_frame

    return filtered_data


def _get_modified_page_figures(
    ctds_filter: List[CallbackTriggerDict],
    ctds_filter_interaction: List[Dict[str, CallbackTriggerDict]],
    ctds_parameters: List[CallbackTriggerDict],
    targets: Optional[List[ModelID]] = None,
) -> Dict[str, Any]:
    if not targets:
        targets = []
    filtered_data = _get_filtered_data(
        targets=targets,
        ctds_filters=ctds_filter,
        ctds_filter_interaction=ctds_filter_interaction,
    )

    parameterized_config = _get_parametrized_config(
        targets=targets,
        parameters=ctds_parameters,
    )

    outputs: Dict[str, Any] = {}
    for target in targets:
        outputs[target] = model_manager[target](data_frame=filtered_data[target], **parameterized_config[target])

    return outputs



import pandas as pd
import pytest

import vizro.models as vm
from vizro.managers import model_manager
from vizro.models._action._actions_chain import ActionsChain
from vizro.models._controls.filter import Filter, _filter_between, _filter_isin
from vizro.models.types import CapturedCallable


class TestFilterFunctions:
    @pytest.mark.parametrize(
        "data, value, expected",
        [
            ([1, 2, 3, 4, 5], [2.5, 3.5], [False, False, True, False, False]),  # Standard test
            ([1, 2, 3, 4, 5], [2, 4], [False, True, True, True, False]),  # Test for inclusive both ends
            ([1, 2, 3, 4, 5], [1, 5], [True, True, True, True, True]),  # Test for inclusive all
            ([1, 2, 3, 4, 5], [4, 2], [False, False, False, False, False]),  # Test for inverted values
            ([], [2, 4], pd.Series([], dtype=bool)),  # Test for empty series
            (
                [1.1, 2.2, 3.3, 4.4, 5.5],
                [2.1, 4.5],
                [False, True, True, True, False],
            ),  # Test with float data
        ],
    )
    def test_filter_between(self, data, value, expected):
        series = pd.Series(data)
        expected = pd.Series(expected)
        result = _filter_between(series, value)
        pd.testing.assert_series_equal(result, expected)

    @pytest.mark.parametrize(
        "data, value, expected",
        [
            ([1, 2, 3, 4, 5], [2, 4], [False, True, False, True, False]),  # Test for integers
            (["apple", "banana", "orange"], ["banana", "grape"], [False, True, False]),  # Test for strings
            (
                [1.1, 2.2, 3.3, 4.4, 5.5],
                [2.2, 4.4],  # Test for float values
                [False, True, False, True, False],
            ),
            ([1, 2, 3, 4, 5], [], [False, False, False, False, False]),  # Test for empty value list
        ],
    )
    def test_filter_isin(self, data, value, expected):
        series = pd.Series(data)
        expected = pd.Series(expected)
        result = _filter_isin(series, value)
        pd.testing.assert_series_equal(result, expected)


@pytest.mark.usefixtures("managers_one_page_two_graphs")
class TestFilterInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_filter_mandatory_only(self):
        filter = Filter(column="foo")
        assert filter.type == "filter"
        assert filter.column == "foo"
        assert filter.targets == []

    def test_create_filter_mandatory_and_optional(self):
        filter = Filter(column="foo", targets=["scatter_chart", "bar_chart"], selector=vm.RadioItems())
        assert filter.type == "filter"
        assert filter.column == "foo"
        assert filter.targets == ["scatter_chart", "bar_chart"]
        assert isinstance(filter.selector, vm.RadioItems)

    def test_check_target_present_valid(self):
        Filter(column="foo", targets=["scatter_chart", "bar_chart"])

    def test_check_target_present_invalid(self):
        with pytest.raises(ValueError, match="Target invalid_target not found in model_manager."):
            Filter(column="foo", targets=["invalid_target"])


class TestPreBuildMethod:
    def test_set_targets_valid(self, managers_one_page_two_graphs):
        # Core of tests is still interface level
        filter = vm.Filter(column="country")
        # Special case - need filter in the context of page in order to run filter.pre_build
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert set(filter.targets) == {"scatter_chart", "bar_chart"}

    def test_set_targets_invalid(self, managers_one_page_two_graphs):
        filter = vm.Filter(column="invalid_choice")
        model_manager["test_page"].controls = [filter]

        with pytest.raises(ValueError, match="Selected column invalid_choice not found in any dataframe on this page."):
            filter.pre_build()

    @pytest.mark.parametrize(
        "test_input,expected", [("country", "categorical"), ("year", "numerical"), ("lifeExp", "numerical")]
    )
    def test_set_column_type(self, test_input, expected, managers_one_page_two_graphs):
        filter = vm.Filter(column=test_input)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter._column_type == expected

    @pytest.mark.parametrize(
        "test_input,expected", [("country", vm.Dropdown), ("year", vm.RangeSlider), ("lifeExp", vm.RangeSlider)]
    )
    def test_set_selector(self, test_input, expected, managers_one_page_two_graphs):
        filter = vm.Filter(column=test_input)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert isinstance(filter.selector, expected)
        assert filter.selector.title == test_input.title()

    @pytest.mark.parametrize("test_input", [vm.Slider(), vm.RangeSlider()])
    def test_set_slider_values_incompatible_column_type(self, test_input, managers_one_page_two_graphs):
        filter = vm.Filter(column="country", selector=test_input)
        model_manager["test_page"].controls = [filter]
        with pytest.raises(
            ValueError,
            match=f"Chosen selector {test_input.type} is not compatible with categorical column '{filter.column}'.",
        ):
            filter.pre_build()

    @pytest.mark.parametrize("test_input", [vm.Slider(), vm.RangeSlider()])
    def test_set_slider_values_shared_column_inconsistent_dtype(
        self, test_input, managers_shared_column_different_dtype
    ):
        filter = vm.Filter(column="shared_column", selector=test_input)
        model_manager["graphs_with_shared_column"].controls = [filter]
        with pytest.raises(
            ValueError,
            match=f"Non-numeric values detected in the shared data column '{filter.column}' for targeted charts. "
            f"Please ensure that the data column contains the same data type across all targeted charts.",
        ):
            filter.pre_build()

    @pytest.mark.parametrize("test_input", [vm.Slider(), vm.RangeSlider()])
    def test_set_slider_values_defaults_min_max_none(self, test_input, gapminder, managers_one_page_two_graphs):
        filter = vm.Filter(column="lifeExp", selector=test_input)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.min == gapminder.lifeExp.min()
        assert filter.selector.max == gapminder.lifeExp.max()

    @pytest.mark.parametrize("test_input", [vm.Slider(min=3, max=5), vm.RangeSlider(min=3, max=5)])
    def test_set_slider_values_defaults_min_max_fix(self, test_input, managers_one_page_two_graphs):
        filter = vm.Filter(column="lifeExp", selector=test_input)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.min == 3
        assert filter.selector.max == 5

    @pytest.mark.parametrize("test_input", [vm.Checklist(), vm.Dropdown(), vm.RadioItems()])
    def test_set_categorical_selectors_options_defaults_options_none(
        self, test_input, gapminder, managers_one_page_two_graphs
    ):
        filter = vm.Filter(column="continent", selector=test_input)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.options == sorted(set(gapminder["continent"]))

    @pytest.mark.parametrize(
        "test_input",
        [
            vm.Checklist(options=["Africa", "Europe"]),
            vm.Dropdown(options=["Africa", "Europe"]),
            vm.RadioItems(options=["Africa", "Europe"]),
        ],
    )
    def test_set_categorical_selectors_options_defaults_options_fix(self, test_input, managers_one_page_two_graphs):
        filter = vm.Filter(column="continent", selector=test_input)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.options == ["Africa", "Europe"]

    @pytest.mark.parametrize("test_input", ["country", "year", "lifeExp"])
    def test_set_actions(self, test_input, managers_one_page_two_graphs):
        filter = vm.Filter(column=test_input)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        default_action = filter.selector.actions[0]
        assert isinstance(default_action, ActionsChain)
        assert isinstance(default_action.actions[0].function, CapturedCallable)
        assert default_action.actions[0].id == f"filter_action_{filter.id}"


@pytest.mark.usefixtures("managers_one_page_two_graphs")
class TestFilterBuild:
    """Tests filter build method."""

    @pytest.mark.parametrize(
        "test_column,test_selector",
        [
            ("continent", vm.Checklist()),
            ("continent", vm.Dropdown()),
            ("continent", vm.RadioItems()),
            ("pop", vm.RangeSlider()),
            ("pop", vm.Slider()),
        ],
    )
    def test_filter_build(self, test_column, test_selector):
        filter = vm.Filter(column=test_column, selector=test_selector)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        result = str(filter.build())
        expected = str(test_selector.build())
        assert result == expected
